# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import List, Optional

from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.oidc.api_clients.oidc_request_auth import (
    prepare_client_secret_auth_payload,
)
from planet_auth.oidc.auth_client import (
    OidcAuthClient,
)
from planet_auth.oidc.auth_client_with_client_pubkey import (
    OidcAuthClientWithClientPubkey,
    OidcAuthClientWithPubKeyClientConfig,
)
from planet_auth.oidc.auth_client_with_client_secret import (
    OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment,
    OidcAuthClientWithClientSecretClientConfig,
)
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.auth_client_default_authenticators import (
    OidcAuthClientWithRefreshOrReloginOidcTokenRequestAuthenticator,
)


class ClientCredentialsClientSecretClientConfig(OidcAuthClientWithClientSecretClientConfig):
    """
    Configuration required for [planet_auth.ClientCredentialsClientSecretAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_client_credentials_secret",
            "auth_client_class": ClientCredentialsClientSecretAuthClient,
            "display_name": "Client Credentials (Client Secret)",
            "description": "OAuth2 Client Credentials Flow using a shared client secret for authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class ClientCredentialsAuthClientBase(
    OidcAuthClientWithRefreshOrReloginOidcTokenRequestAuthenticator, OidcAuthClient, ABC
):
    def can_login_unattended(self) -> bool:
        # A valid config should always have everything we need.
        # What exactly that is will vary between the client auth methods.
        return True


class ClientCredentialsClientSecretAuthClient(
    ClientCredentialsAuthClientBase, OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment
):
    """
    AuthClient implementation that implements the OAuth client credentials grant
    to obtain tokens for the client itself.  This implementation is for confidential
    clients that use a client secret to protect the client confidentiality.
    """

    def __init__(self, client_config: ClientCredentialsClientSecretClientConfig):
        super().__init__(client_config)
        self._ccauth_client_config = client_config

    # With client credentials and a simple client secret, the auth server
    # insists that we put the secret in the payload during the initial
    # token request.  For other commands (e.g. token validate) it permits
    # us to send it an an auth header.  We prefer this as the default since
    # an auth header should be less likely to land in a log than URL
    # parameters or request payloads.
    def _client_auth_enricher_login(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        auth_payload = prepare_client_secret_auth_payload(
            client_id=self._oidc_client_secret_client_config.client_id(),
            client_secret=self._oidc_client_secret_client_config.client_secret(),
        )
        enriched_payload = {**raw_payload, **auth_payload}
        return enriched_payload, None

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[dict],
        **kwargs,
    ) -> FileBackedOidcCredential:
        self._warn_password_kwarg(**kwargs)
        self._warn_ignored_kwargs(["username", "password"], **kwargs)
        return FileBackedOidcCredential(
            self.token_client().get_token_from_client_credentials(
                client_id=self._ccauth_client_config.client_id(),
                requested_scopes=requested_scopes,
                requested_audiences=requested_audiences,
                auth_enricher=self._client_auth_enricher_login,
                extra=extra,
            )
        )


class ClientCredentialsPubKeyClientConfig(OidcAuthClientWithPubKeyClientConfig):
    """
    Configuration required for [planet_auth.ClientCredentialsPubKeyAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_client_credentials_pubkey",
            "auth_client_class": ClientCredentialsPubKeyAuthClient,
            "display_name": "Client Credentials (Public Key)",
            "description": "OAuth2 Client Credentials Flow using public key client authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class ClientCredentialsPubKeyAuthClient(ClientCredentialsAuthClientBase, OidcAuthClientWithClientPubkey):
    """
    AuthClient implementation that implements the OAuth client credentials grant
    to obtain tokens for the client itself.  This implementation is for confidential
    clients that use a public/private keypair to protect the client confidentiality.
    """

    def __init__(self, client_config: ClientCredentialsPubKeyClientConfig):
        super().__init__(client_config)
        self._pubkey_client_config = client_config

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[dict],
        **kwargs,
    ):
        self._warn_password_kwarg(**kwargs)
        self._warn_ignored_kwargs(["username", "password"], **kwargs)
        return FileBackedOidcCredential(
            self.token_client().get_token_from_client_credentials(
                client_id=self._pubkey_client_config.client_id(),
                requested_scopes=requested_scopes,
                requested_audiences=requested_audiences,
                auth_enricher=self._client_auth_enricher,
                extra=extra,
            )
        )



# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json
import pathlib
import sys
import textwrap
import time
import typing

from planet_auth import (
    AuthException,
    TokenValidator,
    OidcMultiIssuerValidator,
)
from planet_auth.util import custom_json_class_dumper

from .options import (
    opt_audience,
    opt_issuer,
    opt_token,
    opt_token_file,
    opt_human_readable,
)
from .util import recast_exceptions_to_click


class _jwt_human_dumps:
    """
    Wrapper object for controlling the json.dumps behavior of JWTs so that
    we can display a version different from what is stored in memory.

    For pretty printing JWTs, we convert timestamps into
    human-readable strings.
    """

    def __init__(self, data):
        self._data = data

    def __json_pretty_dumps__(self):
        def _human_timestamp_iso(d):
            for key, value in list(d.items()):
                if key in ["iat", "exp", "nbf"] and isinstance(value, int):
                    fmt_time = time.strftime("%Y-%m-%dT%H:%M:%S%z", time.localtime(value))
                    if (key == "exp") and (d[key] < time.time()):
                        fmt_time += " (Expired)"
                    d[key] = fmt_time
                elif isinstance(value, dict):
                    _human_timestamp_iso(value)
            return d

        json_dumps = self._data.copy()
        _human_timestamp_iso(json_dumps)
        return json_dumps


def json_dumps_for_jwt_dict(data: dict, human_readable: bool, indent: int = 2):
    if human_readable:
        return json.dumps(_jwt_human_dumps(data), indent=indent, sort_keys=True, default=custom_json_class_dumper)
    else:
        return json.dumps(data, indent=2, sort_keys=True)


def print_jwt_parts(raw, header, body, signature, human_readable):
    if raw:
        print(f"RAW:\n    {raw}\n")

    if header:
        print(
            f'HEADER:\n{textwrap.indent(json_dumps_for_jwt_dict(data=header, human_readable=human_readable), prefix="    ")}\n'
        )

    if body:
        print(
            f'BODY:\n{textwrap.indent(json_dumps_for_jwt_dict(body, human_readable=human_readable), prefix="    ")}\n'
        )

    if signature:
        pretty_hex_signature = ""
        i = 0
        for c in signature:
            if i == 0:
                pass
            elif (i % 16) != 0:
                pretty_hex_signature += ":"
            else:
                pretty_hex_signature += "\n"

            pretty_hex_signature += "{:02x}".format(c)
            i += 1

        print(f'SIGNATURE:\n{textwrap.indent(pretty_hex_signature, prefix="    ")}\n')


def hazmat_print_jwt(token_str, human_readable):
    print("UNTRUSTED JWT Decoding\n")
    if token_str:
        (hazmat_header, hazmat_body, hazmat_signature) = TokenValidator.hazmat_unverified_decode(token_str)
        print_jwt_parts(
            raw=token_str,
            header=hazmat_header,
            body=hazmat_body,
            signature=hazmat_signature,
            human_readable=human_readable,
        )


@click.group("jwt", invoke_without_command=True)
@click.pass_context
def cmd_jwt(ctx):
    """
    JWT utility for working with tokens.  These functions are primarily targeted
    towards debugging usage.  Many of the functions do not perform token validation.
    THE CONTENTS OF UNVALIDATED TOKENS MUST BE TREATED AS UNTRUSTED AND POTENTIALLY
    MALICIOUS.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)


def _get_token_or_fail(token_opt: typing.Optional[str], token_file_opt: typing.Optional[pathlib.Path]):
    if token_opt:
        token = token_opt
    elif token_file_opt:
        with open(token_file_opt, mode="r", encoding="UTF-8") as file_r:
            token = file_r.read()
    else:
        # click.echo(ctx.get_help())
        # click.echo()
        raise click.UsageError("A token must be provided.")
    return token


@cmd_jwt.command("decode")
@click.pass_context
@opt_human_readable
@opt_token
@opt_token_file
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_jwt_decode(ctx, token: str, token_file: pathlib.Path, human_readable):
    """
    Decode a JWT token WITHOUT PERFORMING ANY VALIDATION.
    """
    token_to_print = _get_token_or_fail(token_opt=token, token_file_opt=token_file)
    hazmat_print_jwt(token_str=token_to_print, human_readable=human_readable)


@cmd_jwt.command("validate-oauth")
@click.pass_context
@opt_human_readable
@opt_token
@opt_token_file
@opt_audience()
@opt_issuer()
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_jwt_validate_oauth(ctx, token, token_file, audience, issuer, human_readable):
    """
    Perform signature validation on an RFC 9068 compliant JWT token.
    The `iss` and `aud` claims will be used to look up signing keys
    using OAuth2/OIDC discovery protocols and perform basic validation
    checks.

    This command performs only basic signature verification and token validity
    checks.  For checks against auth server token revocation lists, see the `oauth`
    command.  For deeper checks specific to the claims and structure of
    Identity or Access tokens, see the `oauth` command.

    WARNING:\n
    THIS TOOL IS ABSOLUTELY INAPPROPRIATE FOR PRODUCTION TRUST USAGE.  This is a
    development and debugging utility.  The default behavior to inspect the token
    for issuer and audience information used to validate the token is wholly
    incorrect for a production use case.  The decision of which issuers to
    trust with what audiences MUST be controlled by the service operator.
    """
    token_to_validate = _get_token_or_fail(token_opt=token, token_file_opt=token_file)
    (hazmat_header, hazmat_body, hazmat_signature) = TokenValidator.hazmat_unverified_decode(token_to_validate)

    if issuer:
        validation_iss = issuer
    else:
        if not hazmat_body.get("iss"):
            raise click.BadParameter(
                "The provided token does not contain an `iss` claim.  Is the provided JWT RFC 9068 compliant?"
            )
        validation_iss = hazmat_body.get("iss")

    if audience:
        validation_aud = audience
    else:
        if not hazmat_body.get("aud"):
            raise click.BadParameter(
                "The provided token does not contain an `aud` claim.  Is the provided JWT RFC 9068 compliant?"
            )
        hazmat_aud = hazmat_body.get("aud")
        if isinstance(hazmat_aud, list):
            validation_aud = hazmat_aud[0]
        else:
            validation_aud = hazmat_aud

    validator = OidcMultiIssuerValidator.from_auth_server_urls(
        trusted_auth_server_urls=[validation_iss], audience=validation_aud, log_result=False
    )
    validated_body, _ = validator.validate_access_token(token_to_validate, do_remote_revocation_check=False)
    # Validation throws on error
    click.echo("TOKEN OK")
    print_jwt_parts(
        raw=token_to_validate,
        header=hazmat_header,
        body=validated_body,
        signature=hazmat_signature,
        human_readable=human_readable,
    )


@cmd_jwt.command("validate-rs256", hidden=True)
@click.pass_context
@opt_human_readable
@opt_token
@opt_token_file
@recast_exceptions_to_click(AuthException, FileNotFoundError, NotImplementedError)
def cmd_jwt_validate_rs256(ctx, token, token_file, human_readable):
    """
    Validate a JWT signed with a RS256 signature
    """
    # token_to_validate = _get_token_or_fail(token_opt=token, token_file_opt=token_file)
    raise NotImplementedError("Command not implemented")


@cmd_jwt.command("validate-hs512", hidden=True)
@click.pass_context
@opt_human_readable
@opt_token
@opt_token_file
@recast_exceptions_to_click(AuthException, FileNotFoundError, NotImplementedError)
def cmd_jwt_validate_hs512(ctx, token, token_file, human_readable):
    """
    Validate a JWT signed with a HS512 signature
    """
    # token_to_validate = _get_token_or_fail(token_opt=token, token_file_opt=token_file)
    raise NotImplementedError("Command not implemented")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Optional, Union

import planet_auth.logging.auth_logger
from planet_auth.credential import Credential
from planet_auth.auth_client import AuthClient, AuthClientConfig, AuthClientException
from planet_auth.static_api_key.request_authenticator import FileBackedApiKeyRequestAuthenticator
from planet_auth.static_api_key.static_api_key import FileBackedApiKey

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class StaticApiKeyAuthClientException(AuthClientException):
    def __init__(self, message=None):
        super().__init__(message=message)


class StaticApiKeyAuthClientConfig(AuthClientConfig):
    def __init__(self, api_key=None, bearer_token_prefix="Bearer", **kwargs):
        super().__init__(**kwargs)
        if api_key:
            self._data["api_key"] = api_key
            if bearer_token_prefix:
                # Only save the prefix if we have a key.
                # the config file base class only expects _data to be whole, never partial.
                # (Static API keys auth client is a funny case, where an AuthClientConfig
                # and a Credential object are pretty much the same.)
                self._data["bearer_token_prefix"] = bearer_token_prefix

    def api_key(self) -> str:
        return self.lazy_get("api_key")

    def bearer_token_prefix(self) -> str:
        return self.lazy_get("bearer_token_prefix")

    @classmethod
    def meta(cls):
        return {
            "client_type": "static_apikey",
            "auth_client_class": StaticApiKeyAuthClient,
            "display_name": "API Key",
            "description": "Static API key authentication.",
            "config_hints": [
                {
                    "config_key": "api_key",
                    "config_key_name": "API Key",
                    "config_key_description": "Static API key used to authenticate requests",
                },
                {
                    "config_key": "bearer_token_prefix",
                    "config_key_name": "Bearer Token Prefix",
                    "config_key_description": "Prefix used Authorization headers with the API key",
                },
            ],
        }


class StaticApiKeyAuthClient(AuthClient):
    def __init__(self, client_config: StaticApiKeyAuthClientConfig):
        super().__init__(client_config)
        self._static_api_key_config = client_config

    def login(
        self, allow_open_browser: Optional[bool] = False, allow_tty_prompt: Optional[bool] = False, **kwargs
    ) -> FileBackedApiKey:
        if self._static_api_key_config.api_key():
            return FileBackedApiKey(
                api_key=self._static_api_key_config.api_key(),
                prefix=self._static_api_key_config.bearer_token_prefix(),
                api_key_file=self._static_api_key_config.path(),
            )
        else:
            raise StaticApiKeyAuthClientException(
                message="Cannot return credential object from login() for static API key AuthClient when no API key is configured."
            )

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> FileBackedApiKeyRequestAuthenticator:
        # If an API key has been configured in the client config, use that and ignore
        # any separate credential file.
        storage_provider = self._auth_client_config.storage_provider()
        if self._static_api_key_config.api_key():
            _credential = FileBackedApiKey(
                api_key=self._static_api_key_config.api_key(),
                prefix=self._static_api_key_config.bearer_token_prefix(),
                api_key_file=self._static_api_key_config.path(),
                storage_provider=self._static_api_key_config.storage_provider(),
            )
        else:
            if isinstance(credential, pathlib.Path):
                _credential = FileBackedApiKey(api_key_file=credential, storage_provider=storage_provider)
            elif isinstance(credential, FileBackedApiKey):
                _credential = credential
            elif credential is None:
                # This will be brain-dead until update_credential() or update_credential_data()
                # is called.  This is useful for initializing properly typed credential objects.
                _credential = FileBackedApiKey(storage_provider=storage_provider)
            else:
                raise TypeError(
                    f"{type(self).__name__} does not support {type(credential)} credentials.  Use file path or FileBackedApiKey."
                )

        return FileBackedApiKeyRequestAuthenticator(api_key_credential=_credential)

    def can_login_unattended(self) -> bool:
        # We could enhance login() to prompt for a static API key
        # if it's not in the config, but the current implementation
        # considered it a required field, so this should always be true.
        return True



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import List, Optional

from planet_auth.auth_client import AuthClientConfigException, AuthClientException
from planet_auth.oidc.auth_client import (
    OidcAuthClientConfig,
    OidcAuthClient,
    OidcAuthClientWithNoneClientAuth,
)
from planet_auth.oidc.auth_client_with_client_pubkey import (
    OidcAuthClientWithClientPubkey,
    OidcAuthClientWithPubKeyClientConfig,
)
from planet_auth.oidc.auth_client_with_client_secret import (
    OidcAuthClientWithClientSecretClientConfig,
    OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment,
)
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.auth_client_default_authenticators import (
    OidcAuthClientWithRefreshingOidcTokenRequestAuthenticator,
)
from planet_auth.oidc.util import create_pkce_challenge_verifier_pair


class AuthCodeClientConfig(OidcAuthClientConfig):
    """
    Configuration required for [planet_auth.AuthCodeAuthClient][]
    """

    # TODO: simplify down to just one url, I think
    def __init__(self, redirect_uri: str = None, local_redirect_uri: str = None, **kwargs):
        super().__init__(**kwargs)
        # Redirect URI must match the client config on the OIDC service,
        # which may permit multiple values. We let our the config provide
        # different URLs for use cases where we either expect to configure a
        # redirect service locally, or expect it to be available remotely.
        # If only one is set, set both to the same value.
        if redirect_uri:
            self._data["redirect_uri"] = redirect_uri
        if local_redirect_uri:
            self._data["local_redirect_uri"] = local_redirect_uri

        if redirect_uri and not local_redirect_uri:
            self._data["local_redirect_uri"] = redirect_uri
        if local_redirect_uri and not redirect_uri:
            self._data["redirect_uri"] = local_redirect_uri

    def check_data(self, data):
        super().check_data(data)
        if not (data.get("redirect_uri") or data.get("local_redirect_uri")):
            raise AuthClientConfigException(
                message="A redirect_uri or local_redirect_uri is required for auth code client."
            )

    def local_redirect_uri(self) -> str:
        return self.lazy_get("local_redirect_uri")

    def redirect_uri(self) -> str:
        return self.lazy_get("redirect_uri")

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_auth_code",
            "auth_client_class": AuthCodeAuthClient,
            "display_name": "Auth Code",
            "description": "OAuth2 Auth Code with PKCE Flow for a non-confidential client.",
            "config_hints": super().meta().get("config_hints")
            + [
                {
                    "config_key": "redirect_uri",
                    "config_key_name": "Login redirect URL",
                    "config_key_description": "The callback URL used by the client to receive the results of authentication from the authorization server.",
                },
                {
                    "config_key": "local_redirect_uri",
                    "config_key_name": "Login redirect URL",
                    "config_key_description": "Alternative callback URL used by the client to receive the results of authentication from the authorization server.",
                },
            ],
        }


class AuthCodeAuthClientException(AuthClientException):
    pass


class AuthCodeAuthClientBase(OidcAuthClientWithRefreshingOidcTokenRequestAuthenticator, OidcAuthClient, ABC):
    def __init__(self, client_config: AuthCodeClientConfig):
        super().__init__(client_config)
        self._authcode_client_config = client_config

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[dict],
        **kwargs,
    ) -> FileBackedOidcCredential:
        """
        Obtain tokens from the OIDC auth server using the Auth Code OAuth
        flow with PKCE.  The Auth Code flow is inherently user interactive,
        and one of allow_tty_prompt or allow_open_browser must be true.

        Args:
            allow_open_browser: specify whether login is permitted to open
                a browser window.
            allow_tty_prompt: specify whether login is permitted to request
                input from the terminal.
            requested_scopes: a list of strings specifying the scopes to
                request.
            requested_audiences: a list of strings specifying the audiences
                to request.
            extra: a dict extra data to pass to the authorization server.
        Returns:
            A FileBackedOidcCredential object
        """

        self._warn_password_kwarg(**kwargs)
        self._warn_ignored_kwargs(["username", "password", "client_id", "client_secret"], **kwargs)

        pkce_code_verifier, pkce_code_challenge = create_pkce_challenge_verifier_pair()
        if allow_open_browser:
            redirect_uri = self._authcode_client_config.local_redirect_uri()
            authcode = self.authorization_client().authcode_from_pkce_auth_request_with_browser_and_callback_listener(
                client_id=self._authcode_client_config.client_id(),
                redirect_uri=redirect_uri,
                requested_scopes=requested_scopes,
                requested_audiences=requested_audiences,
                pkce_code_challenge=pkce_code_challenge,
                extra=extra,
            )
        elif allow_tty_prompt:
            redirect_uri = self._authcode_client_config.redirect_uri()
            authcode = self.authorization_client().authcode_from_pkce_auth_request_with_tty_input(
                client_id=self._authcode_client_config.client_id(),
                redirect_uri=redirect_uri,
                requested_scopes=requested_scopes,
                requested_audiences=requested_audiences,
                pkce_code_challenge=pkce_code_challenge,
                extra=extra,
            )
        else:
            raise AuthCodeAuthClientException(
                message="Both browser and terminal input disallowed.  No way to complete user login."
            )

        token_json = self.token_client().get_token_from_code(
            redirect_uri=redirect_uri,
            client_id=self._authcode_client_config.client_id(),
            code=authcode,
            code_verifier=pkce_code_verifier,
            auth_enricher=self._client_auth_enricher,
            # extra=extra,  # passed to Auth, not code redemption.
        )
        return FileBackedOidcCredential(token_json)

    def can_login_unattended(self) -> bool:
        # Always requires user interaction
        return False


class AuthCodeAuthClient(AuthCodeAuthClientBase, OidcAuthClientWithNoneClientAuth):
    """
    AuthClient implementation that implements the OAuth auth code grant with PKCE
    to obtain user tokens.  This implementation is for public clients that cannot
    maintain client confidentiality.
    """

    # The base and mix-ins pretty much do it all.


# client auth via secret or public/private keypair are optional for auth
# code flow clients. They are used when the client is "private"
class AuthCodeWithClientSecretClientConfig(AuthCodeClientConfig, OidcAuthClientWithClientSecretClientConfig):
    """
    Configuration required for [planet_auth.AuthCodeWithClientSecretAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_auth_code_secret",
            "auth_client_class": AuthCodeWithClientSecretAuthClient,
            "display_name": "Auth Code (Client Secret)",
            "description": "OAuth2 Auth Code with PKCE Flow for a confidential client using a shared client secret for authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class AuthCodeWithClientSecretAuthClient(
    AuthCodeAuthClientBase, OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment
):
    """
    AuthClient implementation that implements the OAuth auth code grant with PKCE
    to obtain user tokens.  This implementation is for confidential
    clients that use a client secret to protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: AuthCodeWithClientSecretClientConfig):
        super().__init__(client_config)


class AuthCodeWithPubKeyClientConfig(AuthCodeClientConfig, OidcAuthClientWithPubKeyClientConfig):
    """
    Configuration required for [planet_auth.AuthCodeWithPubKeyAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_auth_code_pubkey",
            "auth_client_class": AuthCodeWithPubKeyAuthClient,
            "display_name": "Auth Code (Public Key)",
            "description": "OAuth2 Auth Code with PKCE Flow for a confidential client using public key authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class AuthCodeWithPubKeyAuthClient(AuthCodeAuthClientBase, OidcAuthClientWithClientPubkey):
    """
    AuthClient implementation that implements the OAuth auth code grant with PKCE
    to obtain user tokens.  This implementation is for confidential
    clients that use a public/private keypair to protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: AuthCodeWithPubKeyClientConfig):
        super().__init__(client_config)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from planet_auth.oidc.api_clients.api_client import OidcApiClient


# class UserinfoApiException(OidcApiClientException):
#    def __init__(self, **kwargs):
#        super().__init__(**kwargs)


class UserinfoApiClient(OidcApiClient):
    """
    Low level network client for the "Userinfo" OAuth2/OIDC
    network endpoint.
    """

    def __init__(self, userinfo_uri: str):
        """
        Create a new token Userinfo API client
        """
        super().__init__(endpoint_uri=userinfo_uri)

    def _checked_userinfo_call(self, access_token: str) -> Dict:
        return self._checked_get_json_response(params=None, request_auth=self.TokenBearerAuth(access_token))

    def userinfo_from_access_token(self, access_token: str) -> Dict:
        """
        Obtain user information from the authorization server for the user who owns
        the presented access token.

        Parameters:
            access_token: User access token.
        """
        return self._checked_userinfo_call(access_token)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

from planet_auth.storage_utils import FileBackedJsonObject, ObjectStorageProvider_KeyType
from planet_auth.constants import USER_CONFIG_FILE
from planet_auth.logging.auth_logger import getAuthLogger


auth_logger = getAuthLogger()


class PlanetAuthUserConfig(FileBackedJsonObject):
    """
    Access a ~/.planet.json file to store preferences that are not
    part of a particular AuthClient configuration profile.
    """

    def __init__(
        self,
        data: Optional[dict] = None,
        # file_path: Optional[pathlib.Path] = None,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ):
        # if file_path is None:
        #     file_path = PlanetAuthUserConfig._default_user_config_obj_path()
        super().__init__(
            data=data,
            file_path=ObjectStorageProvider_KeyType(USER_CONFIG_FILE),
            # storage_provider=storage_provider
        )


class PlanetAuthUserConfigEnhanced(PlanetAuthUserConfig):
    """
    Provide an enhanced view of the `~/.planet.json` configuration file,
    with utilities to consolidate the retrieval of configuration
    parameters user overrides, environment variables, the configuration file,
    or fallback internals defaults.
    """

    def effective_conf_value(
        self,
        config_key: str,
        override_value: Optional[str] = None,
        fallback_value: Optional[str] = None,
        env_var_name: Optional[str] = None,
        use_env: bool = True,
        use_configfile: bool = True,
    ):
        """
        Return the effective configuration value synthesized from a variety of sources,
        with the following priority applied:
            - Explicit values from the user are given the highest priority.
            - Values taken from environment variables are then considered.
            - Values taken from the user's configuration file are then next considered.
              The default user configuration file is `~/.planet.json`.
            - Built-in default values are considered last.
        Parameters:
            config_key:  The configuration key to look up.  By default, this will be used
                both for reading the user's configuration file, and used as the environment
                variable to search for.
            override_value: User provided override value.  If this value is supplied,
                it will be returned.
            fallback_value: A fallback default value to return if the config parameter
                cannot be found in the user's configuration file or the environment, and the
                user did not provide an explicit override value.
            env_var_name: Name of the environment variable to search for a value.
                If not provided, the `config_key` will also be used as the environment
                variable.
            use_env: Boolean controlling whether the environment should be considered.
                By default, the environment will be searched for configuration values,
                but this behavior can be surprised using this parameter.
            use_configfile: Boolean controlling whether the user configuration file should be considered.
                By default, the user configuration file will be searched for configuration values,
                but this behavior can be surprised using this parameter.
        """
        if override_value:
            return override_value

        if use_env:
            if env_var_name:
                _env_var = env_var_name
            else:
                _env_var = config_key
            if os.getenv(_env_var):
                return os.getenv(_env_var)

        if use_configfile:
            try:
                # It's the caller's job to decide when a reload is safe.
                # config_file_value = self.lazy_reload_get(config_key)
                config_file_value = self.lazy_get(config_key)
                if config_file_value:
                    return config_file_value
            except Exception as ex:
                auth_logger.debug(msg=f"{ex}")

        return fallback_value



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from typing import List, Optional, Tuple

from planet_auth import (
    Auth,
    AuthClientConfig,
    AuthException,
    ObjectStorageProvider,
    OidcMultiIssuerValidator,
    PlanetLegacyRequestAuthenticator,
)
from planet_auth.constants import (
    TOKEN_FILE_SOPS,
    TOKEN_FILE_PLAIN,
    AUTH_CONFIG_FILE_SOPS,
    AUTH_CONFIG_FILE_PLAIN,
)

from planet_auth_utils.profile import Profile
from planet_auth_utils.builtins import Builtins
from planet_auth_utils.plauth_user_config import PlanetAuthUserConfigEnhanced
from planet_auth_utils.constants import EnvironmentVariables
from planet_auth.logging.auth_logger import getAuthLogger

auth_logger = getAuthLogger()


class PlanetAuthFactory:
    @staticmethod
    def _token_file_path(profile_name: str, overide_path: Optional[str], save_token_file: bool):
        # The initialized Auth object just uses whether or not a token file path
        # is set to determine whether to use a credential file.  The layering from
        # the sources of config values needs some handholding
        # to set the token file value correctly to account for this.
        if save_token_file:
            return Profile.get_profile_file_path_with_priority(
                filenames=[TOKEN_FILE_SOPS, TOKEN_FILE_PLAIN], profile=profile_name, override_path=overide_path
            )
        else:
            return None

    @staticmethod
    def _auth_client_config_file_path(profile_name: str):
        return Profile.get_profile_file_path_with_priority(
            filenames=[AUTH_CONFIG_FILE_SOPS, AUTH_CONFIG_FILE_PLAIN], profile=profile_name
        )

    @staticmethod
    def _update_saved_profile_config(plauth_context: Auth, storage_provider: Optional[ObjectStorageProvider] = None):
        _profile_name = plauth_context.profile_name()
        if not _profile_name:
            raise MissingArgumentException("A profile name must be provided if persisting the profile configuration")

        plauth_context._auth_client.config().set_path(
            PlanetAuthFactory._auth_client_config_file_path(profile_name=_profile_name)
        )
        plauth_context._auth_client.config().set_storage_provider(storage_provider=storage_provider)
        plauth_context._auth_client.config().save()

    @staticmethod
    def load_auth_client_config_from_profile(
        profile_name: str,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ) -> Tuple[str, AuthClientConfig]:
        """
        Load the auth client config from a profile name.  Both built-in and used defined
        custom profiles are considered.
        """
        # TODO: does not yet support custom storage providers
        if Builtins.is_builtin_profile(profile_name):
            normalized_profile_name = Builtins.dealias_builtin_profile(profile_name)
            auth_client_config = Builtins.load_builtin_auth_client_config(normalized_profile_name)
        else:
            normalized_profile_name = profile_name.lower()
            auth_client_config = Profile.load_auth_client_config(
                profile=normalized_profile_name,
                # storage_provider=storage_provider,
            )

        return normalized_profile_name, auth_client_config

    @staticmethod
    def _init_context_from_profile(
        profile_name: str,
        token_file_opt: Optional[str] = None,
        save_token_file: bool = True,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ) -> Auth:
        normalized_selected_profile, auth_client_config = PlanetAuthFactory.load_auth_client_config_from_profile(
            profile_name=profile_name,
            # storage_provider=storage_provider,
        )

        token_file_path = PlanetAuthFactory._token_file_path(
            profile_name=normalized_selected_profile, overide_path=token_file_opt, save_token_file=save_token_file  # type: ignore
        )

        auth_logger.debug(msg=f"Initializing Auth from profile {normalized_selected_profile}")
        return Auth.initialize_from_config(
            client_config=auth_client_config,
            token_file=token_file_path,
            profile_name=normalized_selected_profile,
        )

    @staticmethod
    def _init_context_from_oauth_svc_account(
        client_id: str,
        client_secret: str,
        token_file_opt: Optional[str] = None,
        save_token_file: bool = True,
        # profile_name: Optional[str] = None, # TODO? Always currently ad-hoc with this initializer's use.
        save_profile_config: bool = False,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ) -> Auth:
        # TODO: support oauth service accounts that use pubkey, and not just client secrets.
        # TODO: Can we handle different trust realms when initializing a M2M client with
        #       just the Client ID and secret? (akin to how Auth0 implements AUTH0_DOMAIN usage.)
        m2m_realm_name = Builtins.builtin_default_profile_name(client_type="oidc_client_credentials_secret")
        base_client_config = Builtins.builtin_profile_auth_client_config_dict(m2m_realm_name)
        constructed_client_config_dict = {
            **base_client_config,
            "client_id": client_id,
            "client_secret": client_secret,
        }

        adhoc_profile_name = f"{m2m_realm_name}-{client_id}"

        token_file_path = PlanetAuthFactory._token_file_path(
            profile_name=adhoc_profile_name, overide_path=token_file_opt, save_token_file=save_token_file
        )

        auth_logger.debug(msg=f"Initializing Auth for service account {m2m_realm_name}:{client_id}")
        plauth_context = Auth.initialize_from_config_dict(
            client_config=constructed_client_config_dict,
            token_file=token_file_path,
            profile_name=adhoc_profile_name,
        )
        if save_profile_config:
            PlanetAuthFactory._update_saved_profile_config(
                plauth_context=plauth_context,
                # storage_provider=storage_provider,  # Not yet supported here
            )

        return plauth_context

    @staticmethod
    def _init_context_from_client_config(
        client_config: dict,
        profile_name: str,
        initial_token_data: Optional[dict] = None,
        save_token_file: bool = True,
        save_profile_config: bool = True,
        storage_provider: Optional[ObjectStorageProvider] = None,
    ) -> Auth:
        token_file_path = PlanetAuthFactory._token_file_path(
            profile_name=profile_name, overide_path=None, save_token_file=save_token_file
        )

        auth_logger.debug(msg="Initializing Auth from provided configuration")
        plauth_context = Auth.initialize_from_config_dict(
            client_config=client_config,
            initial_token_data=initial_token_data,
            token_file=token_file_path,
            profile_name=profile_name,
            storage_provider=storage_provider,
        )
        if save_profile_config:
            PlanetAuthFactory._update_saved_profile_config(
                plauth_context=plauth_context, storage_provider=storage_provider
            )

        return plauth_context

    # @staticmethod
    # def _init_context_from_legacy_username_password_key(
    #     username: str,
    #     password: str,
    #     token_file_opt: Optional[str] = None,
    #     save_token_file: bool = True,
    # ) -> Auth:
    #     pass
    ## Purposefully not supporting this at this time.
    ## This would be a good place to make use of a built-in legacy type profile.

    @staticmethod
    def _init_context_from_api_key(api_key: str) -> Auth:
        ## We used to use built-in profiles that knew the legacy protocol.
        ## But, since that AuthClient largely exists to turn username/password into
        ## API keys (or legacy JWTs), we can bypass it and simply use an API key AuthClient.
        ## This is desirable since we do not have to know about authentication endpoints.
        # selected_profile_name = Builtins.BUILTIN_PROFILE_NAME_LEGACY
        # constructed_client_config_dict = Builtins._builtin.builtin_client_authclient_config_dicts()[
        #     selected_profile_name]
        # token_file_path = None  # Always None in this case. See _token_file_path() above.
        # constructed_client_config_dict["api_key"] = api_key
        # return Auth.initialize_from_config_dict(
        #    client_config=constructed_client_config_dict,
        #    token_file=token_file_path,
        #    profile_name=selected_profile_name,
        # )
        constructed_client_config_dict = {
            "client_type": "static_apikey",
            "api_key": api_key,
            "bearer_token_prefix": PlanetLegacyRequestAuthenticator.TOKEN_PREFIX,
        }
        adhoc_profile_name = "_PL_API_KEY"
        auth_logger.debug(msg="Initializing Auth from API key")
        plauth_context = Auth.initialize_from_config_dict(
            client_config=constructed_client_config_dict,
            token_file=None,
            profile_name=adhoc_profile_name,
        )

        # if save_profile_config:
        #    PlanetAuthFactory._update_saved_profile_config(
        #        plauth_context=plauth_context,
        #        storage_provider=storage_provider)

        return plauth_context

    @staticmethod
    def initialize_auth_client_context(
        auth_profile_opt: Optional[str] = None,
        auth_client_id_opt: Optional[str] = None,
        auth_client_secret_opt: Optional[str] = None,
        auth_api_key_opt: Optional[str] = None,  # Deprecated
        token_file_opt: Optional[str] = None,  # TODO: Remove? but we still depend on it for Planet Legacy use cases.
        # TODO?: initial_token_data: dict = None,
        save_token_file: bool = True,
        save_profile_config: bool = False,
        # Not supporting custom storage providers at this time.
        # The preferred behavior of Profiles with custom storage providers is TBD.
        # storage_provider: Optional[ObjectStorageProvider] = None,
    ) -> Auth:
        """
        Helper function to initialize the Auth context in applications.

        Between built-in profiles to interactively login users, customer or third party
        registered OAuth clients and corresponding custom profiles that may be saved on disk,
        OAuth service account profiles, and static API keys, there are a number of
        ways to configure how an application built with this library should authenticate
        requests made to the service.  Add to this configration may come from explict
        parameters set by the user, environment variables, configuration files, or values
        hard-coded by the application developer, and the number of possibilities rises.

        This helper function is provided to help build applications with a consistent
        user experience when sharing auth context with the CLI.  This function
        does not support using custom storage providers at this time.

        Arguments to this function are taken to be explicitly set by the user or
        application developer, and are given the highest priority.  Internally, the
        priority used for the source of any particular configuration values is, from
        highest to lowest priority, as follows:
            - Arguments to this function.
            - Environment variables.
            - Values from configuration file.
            - Built-in defaults.

        In constructing the returned Auth context, the following priority is applied, from
        highest to lowest:
            - A user selected auth profile, as specified by `auth_profile_opt`. This may either
              specify a built-in profile name, or a fully custom profile defined by files in
              a `~/.planet/<profile name>` directory.
            - A user selected OAuth service account, as specified by `auth_client_id_opt` and `auth_client_secret_opt`.
            - A user specified API key, as specified by `auth_api_key_opt`
            - A user selected auth profile, as determined from either environment variables or config files.
            - A user selected OAuth service account, as determined from either environment variables or config files.
            - A user selected API key, as determined from either environment variables or config files.
            - A built-in default auth profile, which may require interactive user authentication.

        Example:
            ```python
            @click.group(help="my cli main help message")
            @opt_auth_profile
            @opt_auth_client_id
            @opt_auth_client_secret
            @click.pass_context
            def my_cli_main(ctx, auth_profile, auth_client_id, auth_client_secret):
                ctx.ensure_object(dict)
                ctx.obj["AUTH"] = PlanetAuthFactory.initialize_auth_client_context(
                    auth_profile_opt=auth_profile, auth_client_id_opt=auth_client_id, auth_client_secret_opt=auth_client_secret
                )
                # Click program may now use the auth context in all commands...
            ```
        """
        #
        # Initialize from explicit user selected options
        #
        if auth_profile_opt:
            return PlanetAuthFactory._init_context_from_profile(
                profile_name=auth_profile_opt,
                token_file_opt=token_file_opt,
                save_token_file=save_token_file,
            )

        if auth_client_id_opt and auth_client_secret_opt:
            return PlanetAuthFactory._init_context_from_oauth_svc_account(
                client_id=auth_client_id_opt,
                client_secret=auth_client_secret_opt,
                token_file_opt=token_file_opt,
                save_token_file=save_token_file,
                # profile_name="",  # Always ad-hoc in this path
                save_profile_config=save_profile_config,
            )

        if auth_api_key_opt:
            return PlanetAuthFactory._init_context_from_api_key(
                api_key=auth_api_key_opt,
            )

        #
        # Initialize from implicit user selected options (env and config files)
        #
        user_config_file = PlanetAuthUserConfigEnhanced()
        log_fallback_warning = False
        effective_user_selected_profile = user_config_file.effective_conf_value(
            config_key=EnvironmentVariables.AUTH_PROFILE,
            override_value=auth_profile_opt,
        )
        if effective_user_selected_profile:
            try:
                return PlanetAuthFactory._init_context_from_profile(
                    profile_name=effective_user_selected_profile,
                    token_file_opt=token_file_opt,
                    save_token_file=save_token_file,
                )
            except Exception as e:
                auth_logger.warning(
                    msg=f'Unable to initialize user selected profile "{effective_user_selected_profile}".'
                    f' Profile was selected from {EnvironmentVariables.AUTH_PROFILE} environment variable or "{user_config_file.path()}" configuration file.'
                    f" Error: {e}"
                )
                log_fallback_warning = True

        effective_user_selected_client_id = user_config_file.effective_conf_value(
            config_key=EnvironmentVariables.AUTH_CLIENT_ID,
            override_value=auth_client_id_opt,
        )
        effective_user_selected_client_secret = user_config_file.effective_conf_value(
            config_key=EnvironmentVariables.AUTH_CLIENT_SECRET,
            override_value=auth_client_secret_opt,
        )
        if effective_user_selected_client_id and effective_user_selected_client_secret:
            return PlanetAuthFactory._init_context_from_oauth_svc_account(
                client_id=effective_user_selected_client_id,
                client_secret=effective_user_selected_client_secret,
                token_file_opt=token_file_opt,
                save_token_file=save_token_file,
                # profile_name="",  # Always ad-hoc in this path
                save_profile_config=save_profile_config,
            )

        effective_user_selected_api_key = user_config_file.effective_conf_value(
            config_key=EnvironmentVariables.AUTH_API_KEY,
            override_value=auth_api_key_opt,
        )
        if effective_user_selected_api_key:
            return PlanetAuthFactory._init_context_from_api_key(
                api_key=effective_user_selected_api_key,
            )

        effective_user_selected_api_key = user_config_file.effective_conf_value(
            config_key="key",  # For backwards compatibility
            override_value=auth_api_key_opt,
            use_env=False,
        )
        if effective_user_selected_api_key:
            return PlanetAuthFactory._init_context_from_api_key(
                api_key=effective_user_selected_api_key,
            )

        #
        # Fall back to a built-in default configuration when all else fails.
        #
        if log_fallback_warning:
            auth_logger.warning(
                msg=f'Loading built-in profile "{Builtins.builtin_default_profile_name()}" as a fallback.'
            )

        return PlanetAuthFactory._init_context_from_profile(
            profile_name=Builtins.builtin_default_profile_name(),
            token_file_opt=token_file_opt,
            save_token_file=save_token_file,
        )

    @staticmethod
    def initialize_auth_client_context_from_custom_config(
        client_config: dict,
        profile_name: str,
        initial_token_data: Optional[dict] = None,
        save_token_file: bool = True,
        save_profile_config: bool = True,
        storage_provider: Optional[ObjectStorageProvider] = None,
    ) -> Auth:
        """
        Initialize using the provided client config dictionary
        and custom storage provider for session persistence.

        If a storage provider is not provided, a default file
        based storage provider will be used that is interoperable
        with the plauth CLI utility.  If a custom storage provider
        is supplied, sessions will not be visible to the plauth
        CLI tools.
        """
        return PlanetAuthFactory._init_context_from_client_config(
            client_config=client_config,
            profile_name=profile_name,
            initial_token_data=initial_token_data,
            save_token_file=save_token_file,
            save_profile_config=save_profile_config,
            storage_provider=storage_provider,
        )

    @staticmethod
    def initialize_resource_server_validator(
        environment: str,
        trusted_auth_server_configs: Optional[List[dict]] = None,
    ) -> OidcMultiIssuerValidator:
        """
        Create an OIDC multi issuer validator suitable for
        use by a resource server to validate access tokens in the
        specified deployment environment.

        The passed `environment` must be one of `"production"`, `"staging"`,
        or `"custom"`. `environment` is case-insensitive.

        If `"custom"` is selected, trusted_auth_server_configs` must also be specified.
        If custom is not selected, the aforementioned argument will be ignored.
        See `OidcMultiIssuerValidator.from_auth_server_configs` for more info.
        """
        if not environment:
            raise ValueError(f"Passed environment must be one of {Builtins.builtin_environment_names()}.")

        environment = environment.upper()

        if environment not in Builtins.builtin_environment_names():
            raise ValueError(
                f"Passed environment must be one of {Builtins.builtin_environment_names()}. Instead, got: {environment}"
            )

        _builtin_trust_config = Builtins.builtin_environment(environment)

        if _builtin_trust_config is None:
            if trusted_auth_server_configs is None:
                raise MissingArgumentException(
                    "Custom or unknown environment was selected, but trusted_auth_server_configs was not supplied."
                )

            return OidcMultiIssuerValidator.from_auth_server_configs(trusted_auth_server_configs)

        if trusted_auth_server_configs is not None:
            warnings.warn(
                f"Custom environment not selected; trusted_auth_server_configs will be ignored in favor of the built in configuration for {environment}.",
                UserWarning,
            )

        return OidcMultiIssuerValidator.from_auth_server_configs(
            trusted_auth_server_configs=_builtin_trust_config,
        )


class MissingArgumentException(AuthException):
    """Raised when not all custom environment arguments are specified."""



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from planet_auth.storage_utils import FileBackedJsonObject


class Credential(FileBackedJsonObject):
    """
    A storage backed credential.
    A credential is expected to be a json dict.  Per the base class default
    storage provider implementation, clear-text .json files or .sops.json files
    with field level encryption are supported.  Custom storage providers may
    offer different functionality.
    """

    def __init__(self, data=None, file_path=None, storage_provider=None):
        super().__init__(data=data, file_path=file_path, storage_provider=storage_provider)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Package providing an OAuth2/OIDC implementation of the planet_auth package
interfaces.  This is a generic OAuth2/OIDC auth client, and knows nothing
about Planet APIs.

AuthClients are provided for a number of authentication flows, suitable
for user interactive or headless use cases.

Several Request Authenticators are provided that can use the OIDC credentials
obtained from an AuthClient.  Since frequent credential refresh is a part of
using OAuth2/OIDC, these authenticators handle this transparently.
"""



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import os
from typing import List, Optional

from planet_auth import AuthClientConfig
from planet_auth_utils.profile import ProfileException
from planet_auth_utils.constants import EnvironmentVariables
from planet_auth.logging.auth_logger import getAuthLogger
from .builtins_provider import BuiltinConfigurationProviderInterface, EmptyBuiltinProfileConstants

auth_logger = getAuthLogger()


class BuiltinsException(ProfileException):
    pass


def _load_builtins_worker(builtin_provider_fq_class_name, log_warning=False):
    if not builtin_provider_fq_class_name:
        return

    module_name, _, class_name = builtin_provider_fq_class_name.rpartition(".")
    if module_name and class_name:
        try:
            builtin_provider_module = importlib.import_module(module_name)  # nosemgrep - WARNING - See below
            if class_name not in builtin_provider_module.__dict__:
                auth_logger.warning(
                    msg=f"Error loading built-in provider. Module {module_name} does not contain class {class_name}.",
                )
            else:
                provider_instance = builtin_provider_module.__dict__[class_name]()
                return provider_instance
        except ImportError as ie:
            if log_warning:
                auth_logger.warning(
                    msg=f'Error loading built-in provider "{builtin_provider_fq_class_name}". Error: {ie.msg}.',
                )
    else:
        auth_logger.warning(
            msg=f'"{builtin_provider_fq_class_name}" could not be parsed for module and class name while loading built-in provider.',
        )
    return None


def _load_builtins() -> BuiltinConfigurationProviderInterface:
    # Highest priority : injected
    # WARNING: This environment variable is highly sensitive.
    #     Undermining it can undermine client or service security.
    #     It is a convenience for seamless developer experience, but maybe
    #     we should not be so eager to please.
    builtin_provider = _load_builtins_worker(os.getenv(EnvironmentVariables.AUTH_BUILTIN_PROVIDER))
    if builtin_provider:
        return builtin_provider

    # Next priority : Well known implementations (Planet engineering internal)
    builtin_provider = _load_builtins_worker("planet_auth_config.BuiltinConfigurationProvider")
    if builtin_provider:
        return builtin_provider

    # Next priority : Well known implementations (Planet Python SDK)
    builtin_provider = _load_builtins_worker("planet.auth_builtins._BuiltinConfigurationProvider")
    if builtin_provider:
        return builtin_provider

    # Fallback : The empty provider
    return EmptyBuiltinProfileConstants()


class Builtins:
    _builtin: BuiltinConfigurationProviderInterface = None  # type: ignore

    @staticmethod
    def _load_builtin_jit():
        if not Builtins._builtin:
            Builtins._builtin = _load_builtins()

    @staticmethod
    def is_builtin_profile(profile: str) -> bool:
        Builtins._load_builtin_jit()
        if profile:
            _profile = profile.lower()
        else:
            _profile = None

        return (
            _profile in Builtins._builtin.builtin_client_authclient_config_dicts()
            or _profile in Builtins._builtin.builtin_client_profile_aliases()
        )

    @staticmethod
    def is_builtin_profile_alias(profile: str) -> bool:
        Builtins._load_builtin_jit()
        if profile:
            _profile = profile.lower()
        else:
            _profile = None
        return _profile in Builtins._builtin.builtin_client_profile_aliases()

    @staticmethod
    def dealias_builtin_profile(profile: str) -> str:
        Builtins._load_builtin_jit()

        if Builtins.is_builtin_profile(profile):
            _dealiased = profile.lower()
            while Builtins.is_builtin_profile_alias(_dealiased):
                _dealiased = Builtins._builtin.builtin_client_profile_aliases().get(_dealiased)  # type: ignore

            return _dealiased
        else:
            # return None
            raise BuiltinsException(message=f"profile {profile} is not a built-in profile.")

    @staticmethod
    def builtin_profile_names() -> List[str]:
        """
        Return a list of all the built-in profile names.
        """
        Builtins._load_builtin_jit()
        return list(Builtins._builtin.builtin_client_authclient_config_dicts().keys()) + list(
            Builtins._builtin.builtin_client_profile_aliases().keys()
        )

    @staticmethod
    def builtin_profile_auth_client_config_dict(profile: str) -> dict:
        Builtins._load_builtin_jit()
        if not profile:
            raise BuiltinsException(message="profile must be set")

        if Builtins.is_builtin_profile(profile):
            _profile = Builtins.dealias_builtin_profile(profile)
            return Builtins._builtin.builtin_client_authclient_config_dicts().get(_profile)  # type: ignore
        else:
            raise BuiltinsException(message=f"profile {profile} is not a built-in profile.")

    # @staticmethod
    # def builtin_profile_auth_client_config(profile: str):
    #     return AuthClientConfig.from_dict(Builtins._builtin_profile_auth_client_configs.get(profile))

    @staticmethod
    def load_builtin_auth_client_config(profile: str) -> AuthClientConfig:
        Builtins._load_builtin_jit()
        if Builtins.is_builtin_profile(profile):
            auth_logger.debug(
                msg=f'Using built-in "{profile.lower()}" auth client configuration (ignoring on disk config, if it exists)',
            )
            client_config = AuthClientConfig.from_dict(Builtins.builtin_profile_auth_client_config_dict(profile))
        else:
            raise BuiltinsException(message=f"profile {profile} is not a built-in profile.")

        return client_config

    @staticmethod
    def builtin_default_profile_name(client_type: Optional[str] = None) -> str:
        Builtins._load_builtin_jit()
        if client_type in Builtins._builtin.builtin_default_profile_by_client_type():
            return Builtins._builtin.builtin_default_profile_by_client_type()[client_type]
        return Builtins._builtin.builtin_default_profile()

    @staticmethod
    def builtin_environment_names() -> List[str]:
        Builtins._load_builtin_jit()
        return Builtins._builtin.builtin_trust_environment_names()

    @staticmethod
    def builtin_environment(environment: str) -> Optional[List[dict]]:
        Builtins._load_builtin_jit()
        # TODO : distinguish between unknown, and known and set to None (e.g. CUSTOM)
        if environment in Builtins._builtin.builtin_trust_environments():
            _builtin_env = Builtins._builtin.builtin_trust_environments().get(environment.upper())
            return _builtin_env
        else:
            raise BuiltinsException(message=f"environment {environment.upper()} is unknown.")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO: We are not entirely consistent.
#       We use prompt_toolkit for major prompts
#       and click or simple code for others.

import click
from typing import Optional

# from prompt_toolkit.shortcuts import input_dialog, radiolist_dialog, yes_no_dialog

# from planet_auth_utils.builtins import Builtins
from planet_auth_utils.plauth_user_config import PlanetAuthUserConfigEnhanced
from planet_auth_utils.constants import EnvironmentVariables


def prompt_and_change_user_default_profile_if_different(
    candidate_profile_name: str, change_default_selection: Optional[bool] = None
):
    config_file = PlanetAuthUserConfigEnhanced()
    try:
        saved_profile_name = config_file.lazy_get(EnvironmentVariables.AUTH_PROFILE)
    except FileNotFoundError:
        saved_profile_name = None  # config_file.effective_conf_value(EnvironmentVariables.AUTH_PROFILE, fallback_value=Builtins.builtin_default_profile_name())

    if not saved_profile_name:
        # Always write a preference if none is saved.
        # Since CLI options and env vars are higher priority than this file,
        # it should not cause surprises.
        do_change_default = True
    elif change_default_selection is not None:
        do_change_default = change_default_selection
    else:
        do_change_default = False
        if saved_profile_name != candidate_profile_name:
            # do_change_default = yes_no_dialog(
            #     title=f'Change user default {EnvironmentVariables.AUTH_PROFILE} saved in {config_file.path()}?',
            #     text=f'Current value and user default differ.\nDo you want to change the user default {EnvironmentVariables.AUTH_PROFILE} from "{saved_profile_name}" to "{candidate_profile_name}"?',
            #     yes_text="Change",
            #     no_text="Keep",
            # ).run()
            do_change_default = click.confirm(
                text=f'\nCurrent value and user default differ.\nDo you want to change the user default {EnvironmentVariables.AUTH_PROFILE} from "{saved_profile_name}" to "{candidate_profile_name}"? ',
            )

    if do_change_default:
        config_file.update_data({EnvironmentVariables.AUTH_PROFILE: candidate_profile_name})
        config_file.save()

    return do_change_default



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
from abc import ABC
from typing import List, Optional

from planet_auth.auth_client import AuthClientException
from planet_auth.oidc.auth_client import (
    OidcAuthClientConfig,
    OidcAuthClient,
    OidcAuthClientWithNoneClientAuth,
)
from planet_auth.oidc.auth_client_with_client_pubkey import (
    OidcAuthClientWithClientPubkey,
    OidcAuthClientWithPubKeyClientConfig,
)
from planet_auth.oidc.auth_client_with_client_secret import (
    OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment,
    OidcAuthClientWithClientSecretClientConfig,
)
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.auth_client_default_authenticators import (
    OidcAuthClientWithRefreshOrReloginOidcTokenRequestAuthenticator,
)


# Resource owner flows are really not recommended.  We are handling the username/password,
# and there isn't a good reason for that.  Especially in environments with federated identity
# providers, this isn't something either the client, the resource server, or our OAuth service
# needs to know. (Does this even work for federated ID providers that require MFA?)
class ResourceOwnerClientConfig(OidcAuthClientConfig):
    """
    Configuration required for [planet_auth.ResourceOwnerAuthClient][]
    """

    def __init__(self, username=None, password=None, **kwargs):
        super().__init__(**kwargs)
        if username:
            self._data["user_name"] = username
        if password:
            self._data["user_password"] = password

    def user_name(self) -> str:
        return self.lazy_get("user_name")

    def user_password(self) -> str:
        return self.lazy_get("user_password")

    # def check_data(self, data):
    #     super().check_data(data)
    #     # Allow - we can prompt the user.
    #     if not (data.get("user_name") and data.get("user_password)):
    #         raise AuthClientConfigException(message="username and password are both required for resource owner client.")
    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_resource_owner",
            "auth_client_class": ResourceOwnerAuthClient,
            "display_name": "Resource Owner",
            "description": "OAuth2 Resource Owner Flow for a non-confidential client." " (Not recommended)",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class ResourceOwnerAuthClientException(AuthClientException):
    pass


class ResourceOwnerAuthClientBase(
    OidcAuthClientWithRefreshOrReloginOidcTokenRequestAuthenticator, OidcAuthClient, ABC
):
    def __init__(self, client_config: ResourceOwnerClientConfig):
        super().__init__(client_config)
        self._resource_owner_client_config = client_config

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[dict],
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ) -> FileBackedOidcCredential:
        if not username:
            if self._resource_owner_client_config.user_name():
                username = self._resource_owner_client_config.user_name()
            else:
                if allow_tty_prompt:
                    username = input("Email: ")
                else:
                    raise ResourceOwnerAuthClientException(
                        message="Username must be provided when performing non-interactive login"
                    )

        if not password:
            if self._resource_owner_client_config.user_password():
                password = self._resource_owner_client_config.user_password()
            else:
                if allow_tty_prompt:
                    password = getpass.getpass(prompt="Password: ")
                else:
                    raise ResourceOwnerAuthClientException(
                        message="Password must be provided when performing non-interactive login"
                    )

        return FileBackedOidcCredential(
            self.token_client().get_token_from_password(
                client_id=self._resource_owner_client_config.client_id(),
                username=username,
                password=password,
                requested_scopes=requested_scopes,
                requested_audiences=requested_audiences,
                auth_enricher=self._client_auth_enricher,
                extra=extra,
            )
        )

    def can_login_unattended(self) -> bool:
        return bool(self._resource_owner_client_config.user_password()) and bool(
            self._resource_owner_client_config.user_name()
        )


class ResourceOwnerAuthClient(ResourceOwnerAuthClientBase, OidcAuthClientWithNoneClientAuth):
    """
    AuthClient implementation that implements the OAuth password grant
    to obtain user tokens.  This implementation is for public
    clients that cannot maintain client confidentiality.
    """

    # The base and mix-ins pretty much do it all.


class ResourceOwnerWithClientSecretClientConfig(ResourceOwnerClientConfig, OidcAuthClientWithClientSecretClientConfig):
    """
    Configuration required for [planet_auth.ResourceOwnerWithClientSecretAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_resource_owner_secret",
            "auth_client_class": ResourceOwnerWithClientSecretAuthClient,
            "display_name": "Resource Owner (Client Secret)",
            "description": "OAuth2 Resource Owner Flow for a confidential client using a shared client secret for authentication."
            " (Not recommended)",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class ResourceOwnerWithClientSecretAuthClient(
    ResourceOwnerAuthClientBase, OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment
):
    """
    AuthClient implementation that implements the OAuth password grant
    to obtain user tokens.  This implementation is for confidential
    clients that use a client secret to protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: ResourceOwnerWithClientSecretClientConfig):
        super().__init__(client_config)


class ResourceOwnerWithPubKeyClientConfig(ResourceOwnerClientConfig, OidcAuthClientWithPubKeyClientConfig):
    """
    Configuration required for [planet_auth.ResourceOwnerWithPubKeyAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_resource_owner_pubkey",
            "auth_client_class": ResourceOwnerWithPubKeyAuthClient,
            "display_name": "Resource Owner (Public Key)",
            "description": "OAuth2 Resource Owner Flow for a confidential client using public key authentication."
            " (Not recommended)",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class ResourceOwnerWithPubKeyAuthClient(ResourceOwnerAuthClientBase, OidcAuthClientWithClientPubkey):
    """
    AuthClient implementation that implements the OAuth password grant
    to obtain user tokens.  This implementation is for confidential
    clients that use a public/private keypair ti protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: ResourceOwnerWithPubKeyClientConfig):
        super().__init__(client_config)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Optional

from planet_auth.logging.events import AuthEvent


class AuthException(Exception):
    """
    Base Exception class for all exceptions thrown from the planet_auth library.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        inner_exception: Optional[Exception] = None,
        event: Optional[AuthEvent] = None,
        **kwargs,
    ):
        super().__init__(message)
        self._inner_exception = inner_exception
        self._event = event

    def event(self) -> Optional[AuthEvent]:
        return self._event

    @classmethod
    def recast(cls, *exceptions, **params):
        """
        Decorator to recast an exception as the specified exception:
        Example:
            ```python
            @AuthException.recast(Exception)
            @AuthException_SubClass.recast(Some_Specific_Exception)
            def raise_my_exception2():
                raise Some_Specific_Exception()
            ```

        Exercise caution when using multiple re-casts.  Recasting as something
        that is then caught by a decorator further up the stack causes problems:
        Example:
            ```python
            @AuthException.recast(Exception) # Causes a problem, since it will catch
                                             # the results of the recasts below
            @AuthException_SomeSubException.recast(Some_Specific_Exception)
            def raise_my_exception2():
                raise Some_Specific_Exception()
            ```
        """
        if not exceptions:
            exceptions = (Exception,)
        # some_parameter = params.get("some_parameter", some_default)

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    raise cls(message="{} ({})".format(str(e), e.__class__.__name__)) from e

            return wrapper

        return decorator


# TODO: should there be a general authentication/authorization base class?
class InvalidTokenException(AuthException):
    """
    Base class for all exceptions that indicate a failure when
    authenticating a token.  Sub-classes may be used to indicate
    more specific errors.

    This exception should not be used to indicate other error
    conditions that may arise during authentication/authorization.
    """

    def __init__(
        self, event: AuthEvent = AuthEvent.TOKEN_INVALID, jwt_header: dict = None, jwt_body: dict = None, **kwargs
    ):
        super().__init__(event=event, **kwargs)
        self._jwt_header = jwt_header
        self._jwt_body = jwt_body

    def jwt_header(self):
        return self._jwt_header

    def jwt_body(self):
        return self._jwt_body



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod, ABC
from typing import Dict, List, Optional

from planet_auth.auth_client import AuthClientConfig, AuthClient, AuthClientConfigException, AuthClientException
from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.oidc.api_clients.authorization_api_client import AuthorizationApiClient
from planet_auth.oidc.api_clients.device_authorization_api_client import DeviceAuthorizationApiClient
from planet_auth.oidc.api_clients.discovery_api_client import DiscoveryApiClient
from planet_auth.oidc.api_clients.introspect_api_client import IntrospectionApiClient
from planet_auth.oidc.api_clients.jwks_api_client import JwksApiClient
from planet_auth.oidc.api_clients.oidc_request_auth import prepare_client_noauth_auth_payload
from planet_auth.oidc.token_validator import TokenValidator
from planet_auth.oidc.api_clients.revocation_api_client import RevocationApiClient
from planet_auth.oidc.api_clients.userinfo_api_client import UserinfoApiClient
from planet_auth.oidc.api_clients.token_api_client import TokenApiClient
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
import planet_auth.logging.auth_logger


auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class OidcAuthClientConfig(AuthClientConfig, ABC):
    """
    Base config class shared by all OAuth/OIDC auth clients.
    """

    def __init__(
        self,
        auth_server: str = None,
        client_id: str = None,
        audiences: List[str] = None,
        scopes: List[str] = None,
        organization: str = None,  # Auth0 extension
        project_id: str = None,  # Planet Labs extension
        issuer: str = None,
        authorization_endpoint: str = None,
        device_authorization_endpoint: str = None,
        introspection_endpoint: str = None,
        jwks_endpoint: str = None,
        revocation_endpoint: str = None,
        userinfo_endpoint: str = None,
        token_endpoint: str = None,
        authorization_callback_acknowledgement: str = None,
        authorization_callback_acknowledgement_file: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._data["auth_server"] = auth_server
        self._data["authorization_callback_acknowledgement"] = authorization_callback_acknowledgement
        self._data["authorization_callback_acknowledgement_file"] = authorization_callback_acknowledgement_file
        self._data["client_id"] = client_id
        self._data["scopes"] = scopes
        self._data["audiences"] = audiences
        self._data["organization"] = organization
        self._data["project_id"] = project_id
        self._data["issuer"] = issuer

        self._data["authorization_endpoint"] = authorization_endpoint
        self._data["device_authorization_endpoint"] = device_authorization_endpoint
        self._data["introspection_endpoint"] = introspection_endpoint
        self._data["jwks_endpoint"] = jwks_endpoint
        self._data["revocation_endpoint"] = revocation_endpoint
        self._data["userinfo_endpoint"] = userinfo_endpoint
        self._data["token_endpoint"] = token_endpoint

        # Loaded JIT. Not in the serialized self._data representation.
        self._authorization_callback_acknowledgement_data: Optional[str] = None

    def check_data(self, data):
        super().check_data(data)
        if not data.get("auth_server"):
            raise AuthClientConfigException(
                message="auth_server must be configured for {} auth client".format(self.__class__.__name__)
            )
        if not data.get("client_id"):
            raise AuthClientConfigException(
                message="client_id must be configured for {} auth client".format(self.__class__.__name__)
            )

        # Do we actually want this restriction in the auth client, or leave that as a local restriction
        # in the validator call, since that is the only place this is a problem?
        if data.get("audiences"):
            if not isinstance(data.get("audiences"), list):
                raise AuthClientConfigException(
                    message="audiences must be a list type for {} auth client.".format(self.__class__.__name__)
                )
            if len(data.get("audiences")) != 1:
                raise AuthClientConfigException(
                    message="while it is a list type, audiences is only permitted to have one value at this time for {} auth client.".format(
                        self.__class__.__name__
                    )
                )

    def auth_server(self) -> str:
        return self.lazy_get("auth_server")

    def authorization_callback_acknowledgement(self) -> str:
        return self.lazy_get("authorization_callback_acknowledgement")

    def authorization_callback_acknowledgement_data(self) -> Optional[str]:
        # TODO: handle key refresh if the file has changed?
        self._lazy_load_authorization_callback_acknowledgement()
        return self._authorization_callback_acknowledgement_data

    def authorization_callback_acknowledgement_file(self) -> str:
        return self.lazy_get("authorization_callback_acknowledgement_file")

    def client_id(self) -> str:
        return self.lazy_get("client_id")

    def scopes(self) -> List[str]:
        return self.lazy_get("scopes")

    def audiences(self) -> List[str]:
        return self.lazy_get("audiences")

    def organization(self) -> str:
        return self.lazy_get("organization")

    def project_id(self) -> str:
        return self.lazy_get("project_id")

    def issuer(self) -> str:
        return self.lazy_get("issuer")

    def authorization_endpoint(self) -> str:
        return self.lazy_get("authorization_endpoint")

    def device_authorization_endpoint(self) -> str:
        return self.lazy_get("device_authorization_endpoint")

    def introspection_endpoint(self) -> str:
        return self.lazy_get("introspection_endpoint")

    def jwks_endpoint(self) -> str:
        return self.lazy_get("jwks_endpoint")

    def revocation_endpoint(self) -> str:
        return self.lazy_get("revocation_endpoint")

    def userinfo_endpoint(self) -> str:
        return self.lazy_get("userinfo_endpoint")

    def token_endpoint(self) -> str:
        return self.lazy_get("token_endpoint")

    def _lazy_load_authorization_callback_acknowledgement(self):
        # TODO: handle refresh if the file has changed?
        if not self._authorization_callback_acknowledgement_data:
            if self.authorization_callback_acknowledgement():
                self._authorization_callback_acknowledgement_data = self.authorization_callback_acknowledgement()
            else:
                ack_file_path = self.authorization_callback_acknowledgement_file()
                if ack_file_path:
                    with open(ack_file_path, mode="r", encoding="UTF-8") as ack_file:
                        self._authorization_callback_acknowledgement_data = ack_file.read()
        ## Nope.  The lower level libs have a fallback built-in default. "None" is valid here.
        # if not self._authorization_callback_acknowledgement_data:
        #    raise AuthClientConfigException(
        #        message="Authorization callback acknowledgement must be configured for auth client."
        #    )

    @classmethod
    def meta(cls):
        return {
            "config_hints": [
                {
                    "config_key": "auth_server",
                    "config_key_name": "Authorization Server",
                    "config_key_description": "Base URL for the OAuth2/OIDC authorization server",
                    # "config_key_default": "auth server URL",
                },
                {
                    "config_key": "client_id",
                    "config_key_name": "Client ID",
                    "config_key_description": "Unique ID assigned to the client",
                },
            ]
        }


class OidcAuthClient(AuthClient, ABC):
    """
    Base class for AuthClient implementations that implement an OAuth/OIDC
    authentication flow.
    """

    def __init__(self, oidc_client_config: OidcAuthClientConfig):
        super().__init__(oidc_client_config)
        self._oidc_client_config = oidc_client_config
        self.__discovery_client = DiscoveryApiClient(auth_server=self._oidc_client_config.auth_server())
        self.__token_client = None
        self.__authorization_client = None
        self.__device_authorization_client = None
        self.__introspection_client = None
        self.__revocation_client = None
        self.__userinfo_client = None
        self.__jwks_client = None
        self.__token_validator = None
        self.__issuer = None
        # TODO: Need authorization_acknowledgement_body here or from OidcAuthClientConfig
        #        This is something that belongs to the client more than the library.
        #        It's also something that's only relevant in a narrow set of circumstances:
        #        An auth code client that is locally listening for and responding
        #        to a redirect callback from the OAuth authorization endpoint.

    def _discovery(self):
        # We know the internals of the discovery client fetch this
        # JIT and caches
        # TODO: this does OIDC discovery.  Should we fall back to OAuth2
        #  discovery?  This comes down to .well-known/openid-configuration
        #  vs .well-known/oauth-authorization-server
        return self.__discovery_client.discovery()

    def _issuer(self):
        # Issuer is normally expected to be the same as the auth server.
        # we handle them separately to allow discovery under the auth
        # server url to set the issuer string value used in validation.
        # I can't see much use for this other than deployments where
        # proxy redirects may be in play. In such cases, a proxy may own
        # a public "auth server" URL that routes to particular instances,
        # and there is an expectation that "issuer" (used for token
        # validation) may deviate from the URL used to locate the auth server.
        if not self.__issuer:
            if self._oidc_client_config.issuer():
                self.__issuer = self._oidc_client_config.issuer()
            else:
                self.__issuer = self._discovery()["issuer"]
                # TODO: should we fall back to the "auth server" if
                #  discovery fails?  In the wild we've seen oauth servers
                #  that lack discovery endpoints, and it would be bad
                #  user experience to force users to provide both
                #  an auth server and an issuer.
        return self.__issuer

    def _token_validator(self):
        if not self.__token_validator:
            self.__token_validator = TokenValidator(self.jwks_client())
        return self.__token_validator

    def authorization_client(self):
        if not self.__authorization_client:
            if self._oidc_client_config.authorization_endpoint():
                auth_endpoint = self._oidc_client_config.authorization_endpoint()
            else:
                auth_endpoint = self._discovery().get("authorization_endpoint")
            if not auth_endpoint:
                raise AuthClientException(
                    message="Authorization endpoint is not available for the current authorization server"
                )
            self.__authorization_client = AuthorizationApiClient(
                authorization_uri=auth_endpoint,
                authorization_callback_acknowledgement_response_body=self._oidc_client_config.authorization_callback_acknowledgement_data(),
            )
        return self.__authorization_client

    def device_authorization_client(self):
        if not self.__device_authorization_client:
            if self._oidc_client_config.device_authorization_endpoint():
                device_auth_endpoint = self._oidc_client_config.device_authorization_endpoint()
            else:
                device_auth_endpoint = self._discovery()["device_authorization_endpoint"]
            self.__device_authorization_client = DeviceAuthorizationApiClient(
                device_authorization_uri=device_auth_endpoint
            )
        return self.__device_authorization_client

    def introspection_client(self):
        if not self.__introspection_client:
            if self._oidc_client_config.introspection_endpoint():
                introspection_endpoint = self._oidc_client_config.introspection_endpoint()
            else:
                introspection_endpoint = self._discovery().get("introspection_endpoint")
            if not introspection_endpoint:
                raise AuthClientException(
                    message="Token introspection endpoint is not available for the current authorization server"
                )
            self.__introspection_client = IntrospectionApiClient(introspection_endpoint)
        return self.__introspection_client

    def jwks_client(self):
        if not self.__jwks_client:
            if self._oidc_client_config.jwks_endpoint():
                jwks_endpoint = self._oidc_client_config.jwks_endpoint()
            else:
                jwks_endpoint = self._discovery().get("jwks_uri")
            if not jwks_endpoint:
                raise AuthClientException(
                    message="JWKS endpoint is not available for the current authorization server"
                )
            self.__jwks_client = JwksApiClient(jwks_endpoint)
        return self.__jwks_client

    def revocation_client(self):
        if not self.__revocation_client:
            if self._oidc_client_config.revocation_endpoint():
                revocation_endpoint = self._oidc_client_config.revocation_endpoint()
            else:
                revocation_endpoint = self._discovery().get("revocation_endpoint")
            if not revocation_endpoint:
                raise AuthClientException(
                    message="Token revocation endpoint is not available for the current authorization server"
                )
            self.__revocation_client = RevocationApiClient(revocation_endpoint)
        return self.__revocation_client

    def userinfo_client(self):
        if not self.__userinfo_client:
            if self._oidc_client_config.userinfo_endpoint():
                userinfo_endpoint = self._oidc_client_config.userinfo_endpoint()
            else:
                userinfo_endpoint = self._discovery().get("userinfo_endpoint")
            if not userinfo_endpoint:
                raise AuthClientException(
                    message="User information endpoint is not available for the current authorization server"
                )
            self.__userinfo_client = UserinfoApiClient(userinfo_endpoint)
        return self.__userinfo_client

    def token_client(self):
        if not self.__token_client:
            if self._oidc_client_config.token_endpoint():
                token_endpoint = self._oidc_client_config.token_endpoint()
            else:
                token_endpoint = self._discovery().get("token_endpoint")
            if not token_endpoint:
                raise AuthClientException(
                    message="Token endpoint is not available for the current authorization server"
                )
            self.__token_client = TokenApiClient(token_endpoint)
        return self.__token_client

    # Note: I don't really like that auth_client knows about the HTTP-ness,
    #       that's the job of the API client classes to abstract.  It's
    #       difficult to entirely separate these concerns, since the high
    #       level concept of "OIDC client type" imposes itself on some of
    #       the low level protocol interactions.
    @abstractmethod
    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        """
        Some OIDC endpoints require client auth, and how auth is done can
        vary depending on how the OIDC provider is configured to handle the
        particular client and the particular flow of the token grant.

        Auth clients implementation classes must implement a method to enrich
        requests with appropriate authentication either by modifying the
        payload, or providing an AuthBase for the request, which will be used
        where needed before sending requests to the authorization servers.

        If no enrichment is needed or appropriate, implementations should
        return the raw payload unmodified, and None for the AuthBase.

        The signature of this abstract method is expected to adhere to
        the api_client.EnricherFuncType function signature.

        See
        https://developer.okta.com/docs/reference/api/oidc/#client-authentication-methods
        """

    def _apply_config_fallback(self, requested_scopes, requested_audiences, extra):
        if not requested_scopes:
            requested_scopes = self._oidc_client_config.scopes()

        if not requested_audiences:
            requested_audiences = self._oidc_client_config.audiences()

        if not extra:
            extra = {}
        if not extra.get("organization"):
            if self._oidc_client_config.organization():
                extra["organization"] = self._oidc_client_config.organization()
        if not extra.get("project_id"):
            if self._oidc_client_config.project_id():
                extra["project_id"] = self._oidc_client_config.project_id()

        return requested_scopes, requested_audiences, extra

    def login(
        self,
        allow_open_browser: Optional[bool] = False,
        allow_tty_prompt: Optional[bool] = False,
        requested_scopes: Optional[List[str]] = None,
        requested_audiences: Optional[List[str]] = None,
        extra: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Obtain tokens from the OIDC auth server using an appropriate login
        flow. The base implementation here handles common config fallback
        behaviors. Concrete subclasses must implement the flow specific logic
        in a _oidc_flow_login() method.
        Args:
            allow_open_browser: specify whether login is permitted to open
                a browser window.
            allow_tty_prompt: specify whether login is permitted to request
                input from the terminal.
            requested_scopes: a list of strings specifying the scopes to
                request.
            requested_audiences: a list of strings specifying the audiences
                to request.
            extra: a dict extra data to pass to the authorization server.
         Returns:
            A FileBackedOidcCredential object
        """
        final_requested_scopes, final_requested_audiences, final_extra = self._apply_config_fallback(
            requested_scopes=requested_scopes, requested_audiences=requested_audiences, extra=extra
        )

        return self._oidc_flow_login(
            allow_open_browser=allow_open_browser,
            allow_tty_prompt=allow_tty_prompt,
            requested_scopes=final_requested_scopes,
            requested_audiences=final_requested_audiences,
            extra=final_extra,
            **kwargs,
        )

    def _warn_password_kwarg(self, **kwargs):
        """
        Helper function for _oidc_flow_login implementations to offer guidance to users
        and developers when options are unnecessary and will be ignored.
        """
        if "password" in kwargs:
            if kwargs["password"]:
                # Safety check. "password" is a legitimate kwarg for some OAuth flows
                # like Resource Owner Flow.  But, it should never be provided to the client
                # of other flows such as Auth Code or Device Code flows.
                # We could simply ignore it in the kwargs, but it's a good opportunity
                # to improve user or developer security practices.
                warning_msg = (
                    "Supplying your password is not a supported option for the current login process. "
                    "Protect your password. Do not expose your password unnecessarily."
                )
                # If we decide we want to not just warn, but halt user interactive
                # clients, uncomment this:
                # if allow_open_browser or allow_tty_prompt:
                #     raise AuthCodeAuthClientException(message=warning_msg)
                auth_logger.warning(msg=warning_msg)

    def _warn_ignored_kwargs(self, ignore_kws: list, **kwargs):
        """
        Helper function for _oidc_flow_login implementations to offer guidance to users
        and developers when options are unnecessary and will be ignored.  This is mostly
        to steer users away from habitually passing unnecessary arguments. OAuth flows
        behave differently enough that extra data through the generic login() kwargs
        is a problem we can anticipate.
        """
        for ignore_kw in ignore_kws:
            if ignore_kw in kwargs:
                if kwargs[ignore_kw]:
                    auth_logger.debug(
                        msg=f'Ignoring "{ignore_kw}" argument to login. '
                        "It is not used for the current login process."
                    )

    @abstractmethod
    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[Dict],
        **kwargs,
    ) -> FileBackedOidcCredential:
        """
        Abstract method that concrete subclasses must implement
        to complete the login flow and return a valid OIDC Credential.
        """

    def refresh(
        self, refresh_token: str, requested_scopes: List[str] = None, extra: Optional[Dict] = None
    ) -> FileBackedOidcCredential:
        """
        Refresh auth tokens using the provided refresh token
        Args:
            refresh_token: the refresh token to use.
            requested_scopes: a list of strings specifying the scopes to
                request during the token refresh. If not specified, server
                default behavior will apply.
            extra: a dict extra data to pass to the authorization server.
        Returns:
            A FileBackedOidcCredential object
        """
        # Disabled scope fallback for now.
        # The client config default may have more than the user consented to,
        # which will result in a failure.  Introspection could be used, but
        # if this approach is taken, the access token should be inspected.
        # Inspecting refresh tokens shows its full power, whereas inspecting
        # the access token reveals if it was down-scoped for some reason.
        # However, access tokes live a short time, and inspection fails after
        # expiration, whereas refresh tokens may live much longer lives.
        #
        # if not requested_scopes:
        #    requested_scopes = self._oidc_client_config.scopes()

        # TODO: extra from config ??  Similar concerns as above.
        #  The general expectation is that an unadorned refresh
        #  "just works" as expected. It should not change what you have.

        return FileBackedOidcCredential(
            self.token_client().get_token_from_refresh(
                client_id=self._oidc_client_config.client_id(),
                refresh_token=refresh_token,
                requested_scopes=requested_scopes,
                auth_enricher=self._client_auth_enricher,
                extra=extra,
            )
        )

    def validate_access_token_remote(self, access_token: str) -> Dict:
        """
        Validate the access token against the OIDC token introspection endpoint
        Parameters:
            access_token: The access token to validate
        Returns:
            The raw validate json payload
        """
        return self.introspection_client().validate_access_token(access_token, self._client_auth_enricher)

    def validate_access_token_local(
        self, access_token: str, required_audience: str = None, scopes_anyof: list = None
    ) -> Dict:
        # Note:
        #     Tokens may have multiple audiences. When someone requests a
        #     token with multiple audiences, the expectation during login is
        #     that the result has multiple audiences.  However, the underlying
        #     behavior of the jwt.decode() validation appears to be satisfied
        #     if the token under validation has any one of the provided
        #     audiences.  It does not check that all of the provided
        #     audiences are present. Our expectation is aligned with
        #     the former: if multiple audiences are requested for validation,
        #     we insist that they are all present.
        #
        # WARNING:
        #     For the above reason, we are currently only accepting
        #     a single value for the required audience.  If you truly
        #     need to check that a token contains multiple audiences,
        #     you should make multiple calls to this method.
        #     Since a consumer should never really self identify as
        #     multiple audiences, and the same claim can mean different
        #     things to different audiences, this is considered expected
        #     behavior.
        if not required_audience:
            conf_audiences = self._oidc_client_config.audiences()
            if not conf_audiences:
                raise AuthClientException(
                    message="Required audience was not specified either in the auth client config, or provided as an argument to token validation."
                )
            if len(conf_audiences) != 1:
                raise AuthClientException(
                    message="When using the auth client config's audiences as the source for required token audience during validaiton, only one audience may be specified."
                )
            required_audience = conf_audiences[0]

        return self._token_validator().validate_token(
            token_str=access_token,
            issuer=self._issuer(),
            audience=required_audience,
            scopes_anyof=scopes_anyof,
        )

    def validate_id_token_remote(self, id_token: str) -> Dict:
        """
        Validate the ID token against the OIDC token introspection endpoint
        Parameters:
            id_token: ID token to validate
        Returns:
            The raw validated json payload
        """
        return self.introspection_client().validate_id_token(id_token, self._client_auth_enricher)

    def validate_id_token_local(self, id_token: str) -> Dict:
        """
        Validate the ID token locally. A remote connection may still be made
        to obtain signing keys.
        Parameters:
            id_token: ID token to validate
        Returns:
            Upon success, the validated token claims are returned
        """
        return self._token_validator().validate_id_token(
            token_str=id_token, issuer=self._issuer(), client_id=self._oidc_client_config.client_id()
        )

    def validate_refresh_token_remote(self, refresh_token: str) -> Dict:
        """
        Validate the refresh token against the OIDC token introspection
        endpoint
        Parameters:
            refresh_token: Refresh token to validate
        Returns:
            The raw validate json payload
        """
        return self.introspection_client().validate_refresh_token(refresh_token, self._client_auth_enricher)

    def revoke_access_token(self, access_token: str) -> None:
        """
        Revoke the access token with the OIDC provider.
        Parameters:
            access_token: The access token to revoke
        """
        self.revocation_client().revoke_access_token(access_token, self._client_auth_enricher)

    def revoke_refresh_token(self, refresh_token: str) -> None:
        """
        Revoke the refresh token with the OIDC provider.
        Parameters:
            refresh_token: The refresh token to revoke
        """
        self.revocation_client().revoke_refresh_token(refresh_token, self._client_auth_enricher)

    def userinfo_from_access_token(self, access_token: str) -> Dict:
        """
        Look up user information from the auth server using the access token.
        Parameters:
            access_token: User access token.
        """
        return self.userinfo_client().userinfo_from_access_token(access_token=access_token)

    def oidc_discovery(self) -> Dict:
        """
        Query the authorization server's OIDC discovery endpoint for server information.
        Returns:
            Returns the OIDC discovery dictionary.
        """
        return self._discovery()

    def get_scopes(self) -> List[str]:
        """
        Query the authorization server for a list of scopes.
        Returns:
            Returns a list of scopes that may be requested during a call
                to login or refresh
        """
        return self._discovery()["scopes_supported"]


class OidcAuthClientWithNoneClientAuth(OidcAuthClient, ABC):
    """
    Mix-in base class for "public" (non-confidential) OAuth/OIDC auth clients.
    """

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        auth_payload = prepare_client_noauth_auth_payload(client_id=self._oidc_client_config.client_id())
        enriched_payload = {**raw_payload, **auth_payload}
        return enriched_payload, None



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from abc import ABC
from typing import Union

from planet_auth.credential import Credential
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.request_authenticator import (
    RefreshOrReloginOidcTokenRequestAuthenticator,
    RefreshingOidcTokenRequestAuthenticator,
)
from planet_auth.oidc.auth_client import OidcAuthClient


def _common_prepare_oidc_credential(self: OidcAuthClient, credential: Union[pathlib.Path, Credential]):
    storage_provider = self._auth_client_config.storage_provider()
    if isinstance(credential, pathlib.Path):
        _credential = FileBackedOidcCredential(credential_file=credential, storage_provider=storage_provider)
    elif isinstance(credential, FileBackedOidcCredential):
        _credential = credential
    elif credential is None:
        # An empty, path-less (in memory) credential to start the request authenticator off with.
        # Authenticators are permitted to obtain credentials JIT. (E.g. Client Credentials Grant.)
        # Otherwise, the request authenticator will be brain-dead until update_credential()
        # or update_credential_data() is called.  This is still useful for initializing properly
        # typed credential objects.
        _credential = FileBackedOidcCredential(storage_provider=storage_provider)
    else:
        raise TypeError(
            f"{type(self).__name__} does not support {type(credential)} credentials.  Use file path or FileBackedOidcCredential."
        )
    return _credential


class OidcAuthClientWithRefreshOrReloginOidcTokenRequestAuthenticator(ABC):
    """
    Mix-in base class for flows that should default to the
    RefreshOrReloginOidcTokenRequestAuthenticator request authenticator.
    """

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> RefreshOrReloginOidcTokenRequestAuthenticator:
        _credential = _common_prepare_oidc_credential(self, credential)  # type: ignore
        return RefreshOrReloginOidcTokenRequestAuthenticator(credential=_credential, auth_client=self)  # type: ignore


class OidcAuthClientWithRefreshingOidcTokenRequestAuthenticator(ABC):
    """
    Mix-in base class for flows that should default to the
    RefreshingOidcTokenRequestAuthenticator request authenticator.
    """

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> RefreshingOidcTokenRequestAuthenticator:
        _credential = _common_prepare_oidc_credential(self, credential)  # type: ignore
        return RefreshingOidcTokenRequestAuthenticator(credential=_credential, auth_client=self)  # type: ignore



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from planet_auth.auth_exception import InvalidTokenException
from planet_auth.logging.events import AuthEvent
from planet_auth.oidc.api_clients.api_client import (
    OidcApiClient,
    OidcApiClientException,
    EnricherFuncType,
    _RequestAuthType,
    _RequestParamsType,
)


class IntrospectionApiException(OidcApiClientException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class IntrospectionRejectionTokenException(InvalidTokenException):
    def __init__(self, event: AuthEvent = AuthEvent.TOKEN_INVALID_INTROSPECTION_REJECTION, **kwargs):
        super().__init__(event=event, **kwargs)


class IntrospectionApiClient(OidcApiClient):
    """
    Low level network client for the "introspection" OAuth2/OIDC
    network endpoint.  Responses are defined by RFC 7662.

    All invalid responses or responses that indicate that a token
    is not active will result in an exception.
    """

    def __init__(self, introspect_uri: str):
        """
        Create a new token introspection API client
        """
        super().__init__(endpoint_uri=introspect_uri)

    @staticmethod
    def _check_introspection_response(json_response: Dict) -> Dict:
        if not isinstance(json_response.get("active"), bool):
            raise IntrospectionApiException(
                message="Unrecognized response: 'active' field was not present or was not boolean in the response."
            )
        # This is a valid response for an invalid token.
        # The decision to throw mirrors the behavior of PyJWT validation failures.
        if not json_response.get("active"):
            raise IntrospectionRejectionTokenException(message="Token is not active according to issuer introspection")
        return json_response

    @staticmethod
    def check_introspection_response(json_response: Dict) -> Dict:
        # Like above, but for external utility consumption
        if not json_response:
            raise IntrospectionApiException(message="Invalid response: introspection result cannot be empty.")
        return IntrospectionApiClient._check_introspection_response(json_response)

    def _checked_introspection_call(
        self, validate_params: _RequestParamsType, auth: Optional[_RequestAuthType]
    ) -> Dict:
        json_response = self._checked_post_json_response(validate_params, auth)
        return self._check_introspection_response(json_response)

    def _validate_token(self, token: str, token_hint: str, auth_enricher: Optional[EnricherFuncType] = None) -> Dict:
        params = {
            "token": token,
            "token_type_hint": token_hint,
        }
        request_auth = None
        if auth_enricher:
            params, request_auth = auth_enricher(params, self._endpoint_uri)
        return self._checked_introspection_call(params, request_auth)

    def validate_access_token(self, access_token: str, auth_enricher: Optional[EnricherFuncType] = None) -> Dict:
        """
        Validate an access token against the OAuth introspection endpoint.
        Invalid tokens will result in an exception.
        """
        return self._validate_token(access_token, "access_token", auth_enricher)

    def validate_id_token(self, id_token: str, auth_enricher: Optional[EnricherFuncType] = None) -> Dict:
        """
        Validate an ID token against the OAuth introspection endpoint.
        Invalid tokens will result in an exception.
        """
        return self._validate_token(id_token, "id_token", auth_enricher)

    def validate_refresh_token(self, refresh_token: str, auth_enricher: Optional[EnricherFuncType] = None) -> Dict:
        """
        Validate a refresh token against the OAuth introspection endpoint.
        Invalid tokens will result in an exception.
        """
        return self._validate_token(refresh_token, "refresh_token", auth_enricher)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jwt
import time
from typing import Dict, List

import planet_auth.logging.auth_logger
from planet_auth.auth_exception import AuthException, InvalidTokenException
from planet_auth.logging.events import AuthEvent
from planet_auth.oidc.api_clients.jwks_api_client import JwksApiClient

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class TokenValidatorException(AuthException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class InvalidArgumentException(TokenValidatorException):
    pass


class ExpiredTokenException(InvalidTokenException):
    def __init__(self, event: AuthEvent = AuthEvent.TOKEN_INVALID_EXPIRED, **kwargs):
        super().__init__(event=event, **kwargs)


class InvalidAlgorithmTokenException(InvalidTokenException):
    def __init__(self, event: AuthEvent = AuthEvent.TOKEN_INVALID_BAD_ALGORITHM, **kwargs):
        super().__init__(event=event, **kwargs)


class ScopeNotGrantedTokenException(InvalidTokenException):
    def __init__(self, event: AuthEvent = AuthEvent.TOKEN_INVALID_BAD_SCOPE, **kwargs):
        super().__init__(event=event, **kwargs)


class UnknownSigningKeyTokenException(InvalidTokenException):
    def __init__(self, event: AuthEvent = AuthEvent.TOKEN_INVALID_BAD_SIGNING_KEY, **kwargs):
        super().__init__(event=event, **kwargs)


_SCOPE_CLAIM_RFC8693 = "scope"  # See RFC 8693, 9068
_SCOPE_CLAIM_OKTA = "scp"  # Okta uses a different claim

DEFAULT_TRUSTED_ALGORITHMS = ["rs256"]


class TokenValidator:
    """
    Helper class to perform validation of OAuth2 JWT tokens.
    This class provides the implementation for the locally performed
    validation in [planet_auth.OidcAuthClient][] classes, but may
    also be used as a stand-alone validator.

    This is a lower level utility class, and has no dependency
    or knowledge of the higher level [planet_auth.AuthClient][]
    or [planet_auth.Auth][] classes.

    For a higher level application utility, see [planet_auth.OidcMultiIssuerValidator][]
    """

    # Keys really don't change often... Maybe quarterly. In all but
    # emergency situations a new key would be published long before the
    # old one is removed from circulation.  We throttle key fetches so we
    # don't degenerate into DOS'ing the JWKS endpoint if presented with
    # requests using tokens with invalid keys.  So, the default fetch
    # interval can be wide.
    # TODO: implement a max interval - need to push out removed keys.
    def __init__(self, jwks_client: JwksApiClient, min_jwks_fetch_interval=300, trusted_algorithms: List[str] = None):
        self._jwks_client = jwks_client
        self._keys_by_id: Dict[str, jwt.PyJWK] = {}
        self._load_time = 0
        self.min_jwks_fetch_interval = min_jwks_fetch_interval
        # TODO: Some guardrails?
        #       Mixing trust of symmetric and asymmetric key algorithms can be dangerous.
        #       This class is really only designed for asymmetric keys. You would not
        #       usually fetch symmetric keys from a jwks endpoint.
        self._trusted_algorithms = set()
        if not trusted_algorithms:
            self._trusted_algorithms.update(DEFAULT_TRUSTED_ALGORITHMS)
        else:
            for alg in trusted_algorithms:
                self._trusted_algorithms.add(alg.lower())

    def _update(self):
        # TODO: bootstrap from cached values? the concern is large
        #   worker pools on disposable nodes generating load on churn.
        jwks_keys = self._jwks_client.jwks_keys()
        new_keys_by_id = {}
        for json_key in jwks_keys:
            # new_keys_by_id[json_key['kid']] = json_key
            try:
                new_keys_by_id[json_key["kid"]] = jwt.PyJWK(jwk_data=json_key)
            except jwt.PyJWKError as pyjwke:
                auth_logger.warning(
                    msg="Error while loading key from JWKS endpoint. Any attempt to verify tokens signed with this key will fail. Error: {}".format(
                        str(pyjwke)
                    ),
                    exception=pyjwke,
                )

        self._keys_by_id = new_keys_by_id
        self._load_time = int(time.time())

    def get_signing_key_by_id(self, key_id):
        key = self._keys_by_id.get(key_id)
        if (not key) and ((self._load_time + self.min_jwks_fetch_interval) < int(time.time())):
            self._update()
            key = self._keys_by_id.get(key_id)
        if not key:
            raise UnknownSigningKeyTokenException(
                message="Could not find signing key for key ID {}".format(key_id),
            )

        return key

    def _get_trusted_algorithm(self, unverified_header):
        # TODO - Can I just remove this and lean on jwt.decode(algorithms=[]) ??
        algorithm = unverified_header.get("alg")
        # Don't trust straight pass-through, since "none" is a valid
        # algorithm. Only trust specific algorithms.
        if not (algorithm and (algorithm.lower() in self._trusted_algorithms)):
            raise InvalidAlgorithmTokenException(message="Unknown or unsupported token algorithm {}".format(algorithm))

        return algorithm

    @TokenValidatorException.recast(jwt.PyJWTError)
    @InvalidTokenException.recast(jwt.InvalidTokenError)
    @ExpiredTokenException.recast(jwt.ExpiredSignatureError)
    def validate_token(
        self,
        token_str: str,
        issuer: str,
        audience: str,
        required_claims: list = None,
        scopes_anyof: list = None,
        nonce: str = None,
    ):
        """
        Validate the provided token string.  Required claims are validated
        for their presence only. It is up to the application to assert
        that the claim values meet thr requirements of the caller's use case.

        If a nonce is provided, the validator will require that the token
        have a nonce claim, and that its value matches the supplied value.

        Parameters:
            token_str: Raw encoded JWT string.
            issuer: Required token issuer.
            audience: Required token audience.
            required_claims: Optional list of additional claims that must be present in the token.
                these claims are only checked for the presence. It is up to the caller to assert
                that the values are appropriate for the application
            scopes_anyof: Optional list of OAuth2 scopes to check for in the token.
                This list is an "any of" list of scopes. As long as one of the scopes in the
                list is present, validation will pass. If none of the scopes are present,
                validation will fail.
            nonce: Optional nonce value to check for in the token.

        Note: A note on scope validation.
              "Any of" scope enforcement semantics may not make sense for all applications.
              "All of" is a feature that might be desirable, but has not been implemented.
              ("Any of" or "all of" for any other arbitrary claim might also have value. But again,
              this is not implemented here. At some point deriving application meaning from the
              claims present is a concern of the higher level application, and not a concern
              of the token validator.)
        """
        # PyJWT should enforce this, but we have unit tests in case...
        if not token_str:
            raise InvalidArgumentException(message="Cannot validate empty string as a token")
        if not issuer:
            # PyJWT does not seem to raise if the issuer is explicitly None, even when
            # verify_iss was selected.
            raise InvalidArgumentException(message="Cannot validate token with no required issuer provided")
        if not audience:
            raise InvalidArgumentException(message="Cannot validate token with no required audience provided")

        unverified_header = jwt.get_unverified_header(token_str)
        trusted_token_algorithm = self._get_trusted_algorithm(unverified_header)
        key_id = unverified_header.get("kid")
        signing_key = self.get_signing_key_by_id(key_id)
        validation_required_claims = ["aud", "exp", "iss"]
        if required_claims:
            validation_required_claims.extend(required_claims)
        if nonce:
            validation_required_claims.append("nonce")
        validated_claims = jwt.decode(  # Throws when invalid.
            token_str,
            signing_key.key,
            algorithms=[trusted_token_algorithm],
            issuer=issuer,
            audience=audience,
            options={
                "require": validation_required_claims,
                "verify_aud": True,
                "verify_exp": True,
                "verify_iss": True,
                "verify_signature": True,
            },
        )
        if nonce:
            if nonce != validated_claims.get("nonce"):
                raise InvalidTokenException(
                    message="Token nonce did not match expected value",
                    jwt_body=validated_claims,
                    event=AuthEvent.TOKEN_INVALID_BAD_NONCE,
                )

        # We check and throw the more specific errors around scope deliberately last.
        # Invalid tokens never get this far.  To get this far, all the basics of authentication
        # have been passed, and it's a question of the proper grants not having been given.
        if scopes_anyof:
            if validated_claims.get(_SCOPE_CLAIM_RFC8693):
                # RFC 8693 places scopes in a space delimited string.
                token_scopes = validated_claims.get(_SCOPE_CLAIM_RFC8693).split()
            elif validated_claims.get(_SCOPE_CLAIM_OKTA):
                # No split.  Okta places a list of strings in the token.
                token_scopes = validated_claims.get(_SCOPE_CLAIM_OKTA)
            else:
                raise InvalidTokenException(
                    message="No OAuth2 Scopes claim could be found in the access token",
                    jwt_body=validated_claims,
                    event=AuthEvent.TOKEN_INVALID_BAD_SCOPE,
                )

            if not list(set(scopes_anyof) & set(token_scopes)):
                raise ScopeNotGrantedTokenException(
                    message="Access token did not grant sufficient scope. One of [{}] required, but [{}] was granted".format(
                        ", ".join(scopes_anyof), ", ".join(token_scopes)
                    ),
                    jwt_body=validated_claims,
                    event=AuthEvent.TOKEN_INVALID_BAD_SCOPE,
                )

        if not validated_claims:
            # If we got here, jwt.decode passed, and we should at a minimum
            # have claims for audience and issuer.
            raise TokenValidatorException(
                message="No claims could be validated, but no error was detected. This should never happen.",
                event=AuthEvent.TRACE,
            )

        return validated_claims

    @staticmethod
    def hazmat_unverified_decode(token_str):
        # WARNING: Treat unverified token claims like toxic waste.
        #          Nothing can be trusted until the token is verified.
        unverified_complete = jwt.decode_complete(token_str, options={"verify_signature": False})  # nosemgrep
        return unverified_complete["header"], unverified_complete["payload"], unverified_complete["signature"]

    # TODO: should we error if the token has a nonce, and none was given to
    #       verify?
    def validate_id_token(self, token_str, issuer, client_id, required_claims=None, nonce=None):
        """
        Validate a JWT this is a OIDC ID token. Steps over and
        above basic token validation are performed, as described in
        https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation
        """
        validated_claims = self.validate_token(
            token_str=token_str, issuer=issuer, audience=client_id, required_claims=required_claims, nonce=nonce
        )

        # If the aud contains multiple values, azp must be present.
        validated_azp = None
        if isinstance(validated_claims.get("aud"), list):
            validated_azp = validated_claims.get("azp")
            if not validated_azp:
                raise InvalidTokenException(
                    message='"azp" claim mut be present when ID token contains' " multiple audiences."
                )

        # if the azp claim is present, it must equal the client ID.
        if validated_azp:
            if validated_azp != client_id:
                raise InvalidTokenException(
                    message='ID token "azp" claim expected to match the client'
                    ' ID "{}", but was "{}"'.format(client_id, validated_azp)
                )

        return validated_claims



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some OIDC endpoints require client auth, and how auth is done can very
depending on how the OIDC provider is configured to handle the particular
client. See
https://developer.okta.com/docs/reference/api/oidc/#client-authentication-methods

This module provides some helper functions for wrangling the requests for
various OIDC client auth methods.  It is for the caller to decide the
appropriate use of these.
"""

import jwt
import time
import uuid

from typing import Dict
from requests.auth import HTTPBasicAuth

from planet_auth.oidc.api_clients.api_client import _RequestAuthType, _RequestParamsType


def _prepare_oidc_client_jwt_payload(audience: str, client_id: str, ttl: int):
    # See
    # https://developer.okta.com/docs/reference/api/oidc/#token-claims-for-client-authentication-with-client-secret-or-private-key-jwt
    # https://datatracker.ietf.org/doc/html/rfc7523
    now = int(time.time())
    unsigned_jwt = {
        "iss": client_id,
        "sub": client_id,
        "aud": audience,
        "iat": now,
        "nbf": now,
        "exp": now + ttl,
        "jti": str(uuid.uuid4()),
    }
    return unsigned_jwt


def prepare_oidc_client_private_key_jwt(audience: str, client_id: str, private_key, ttl: int):
    unsigned_jwt = _prepare_oidc_client_jwt_payload(audience=audience, client_id=client_id, ttl=ttl)
    signed_jwt = jwt.encode(unsigned_jwt, private_key, algorithm="RS256")
    return signed_jwt


def prepare_client_noauth_auth_payload(client_id: str) -> Dict:
    client_secret_auth_payload = {
        "client_id": client_id,
    }
    return client_secret_auth_payload


def prepare_client_secret_request_auth(client_id: str, client_secret: str) -> _RequestAuthType:
    return HTTPBasicAuth(client_id, client_secret)


def prepare_client_secret_auth_payload(client_id: str, client_secret: str) -> _RequestParamsType:
    client_secret_auth_payload = {"client_id": client_id, "client_secret": client_secret}
    return client_secret_auth_payload


def prepare_private_key_assertion_auth_payload(
    audience: str, client_id: str, private_key, ttl: int
) -> _RequestParamsType:
    signed_jwt = prepare_oidc_client_private_key_jwt(
        audience=audience, client_id=client_id, private_key=private_key, ttl=ttl
    )
    assertion_auth_payload = {
        "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
        "client_assertion": signed_jwt,
    }
    return assertion_auth_payload



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations  # https://stackoverflow.com/a/33533514

import pathlib
from typing import Optional, Union

from planet_auth.auth_client import AuthClient, AuthClientConfig
from planet_auth.credential import Credential
from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.storage_utils import ObjectStorageProvider
from planet_auth.logging.auth_logger import getAuthLogger

auth_logger = getAuthLogger()

# class AuthClientContextException(AuthException):
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)


class Auth:
    """
    A container class for initializing and managing a working set of
    authentication objects.  See factory methods and the
    `planet_auth_utils.PlanetAuthFactory` class for user-friendly
    initialization.

    This container class provides an "auth context" is geared toward client
    use cases - that is, authenticating ourselves as a client with the goal
    of making authenticated network API calls to resource servers.  Resource
    servers that have to validate incoming clients' credentials have some different
    concerns not handled here.
    """

    def __init__(
        self,
        auth_client: AuthClient,
        request_authenticator: CredentialRequestAuthenticator,
        token_file_path: Optional[pathlib.Path] = None,
        profile_name: Optional[str] = None,
    ):
        """
        Create a new auth container object with the specified auth components.
        Users should use one of the more friendly static initializer methods.
        """
        auth_logger.debug(
            msg=f"Initializing Auth Context. Profile: {profile_name} ; Type: {type(auth_client).__name__} ; Token file: {token_file_path}"
        )

        self._auth_client = auth_client
        self._request_authenticator = request_authenticator
        self._token_file_path = token_file_path
        self._profile_name = profile_name
        self._storage_provider = auth_client.config().storage_provider()
        # We do not store the credential since implementations are
        # free to change it out from underneath us during operation.
        # This is common with refresh tokens, for example.
        # self._credential = credential

    def auth_client(self) -> AuthClient:
        """
        Get the currently configured auth client.
        Returns:
            The current AuthClient instance.
        """
        return self._auth_client

    def request_authenticator(self) -> CredentialRequestAuthenticator:
        """
        Get the current request authenticator.
        Returns:
            The current CredentialRequestAuthenticator instance.
        """
        return self._request_authenticator

    def token_file_path(self) -> Optional[pathlib.Path]:
        """
        Get the path to the current credentials file.
        Returns:
            The path to the current credentials file.
        """
        return self._token_file_path

    def profile_name(self) -> Optional[str]:
        """
        Get the current profile name, if known. May be empty if initialization
        was performed without using a profile.
        Returns:
            Returns the name of the current profile, if known.
        """
        return self._profile_name

    def request_authenticator_is_ready(self) -> bool:
        """
        Check whether the current context's request_authenticator is ready
        for use.

        A context is considered initialized if it can be used to make
        authenticated client requests without user intervention.  What exactly
        that requires can vary by auth client configured in the context.
        For example, simple API key clients only need an API key in their
        configuration.  OAuth2 user clients need to have performed
        a user login and obtained access or refresh tokens.
        """
        return self._request_authenticator.is_initialized() or self._auth_client.can_login_unattended()

    def login(self, **kwargs) -> Credential:
        """
        Perform a login with the configured auth client.
        This higher level function will ensure that the token is saved to
        storage if Auth context has been configured with a suitable token
        storage path.  Otherwise, the token will be held only in memory.
        In all cases, the request authenticator will also be updated with
        the new credentials.
        """
        new_credential = self._auth_client.login(**kwargs)
        new_credential.set_path(self._token_file_path)
        new_credential.set_storage_provider(self._storage_provider)
        new_credential.save()
        self._request_authenticator.update_credential(new_credential)
        return new_credential

    def device_login_initiate(self, **kwargs) -> dict:
        """
        Initiate the process to login a device with limited UI capabilities.
        The returned dictionary should contain information for the application
        to present to the user, allowing the user to complete the login process
        asynchronously.

        After prompting the user, `device_login_complete()` should be called
        with the same dictionary that was returned by this call.
        """
        return self._auth_client.device_login_initiate(**kwargs)

    def device_login_complete(self, initiated_login_data: dict) -> Credential:
        """
        Complete a login process that was initiated by a call to `device_login_initiate()`.
        This call will perform all the same actions as `login()`.
        """
        new_credential = self._auth_client.device_login_complete(initiated_login_data)
        new_credential.set_path(self._token_file_path)
        new_credential.set_storage_provider(self._storage_provider)
        new_credential.save()
        self._request_authenticator.update_credential(new_credential)
        return new_credential

    # def refresh(self, **kwargs):
    #     pass

    @staticmethod
    def initialize_from_client(
        auth_client: AuthClient,
        initial_token_data: Optional[dict] = None,
        token_file: Optional[Union[str, pathlib.PurePath]] = None,
        profile_name: Optional[str] = None,
    ) -> Auth:
        """
        Construct and initialize a working set of authentication primitives,
        returning them in a new container object.
        Parameters:
            auth_client: An already constructed [AuthClient][planet_auth.AuthClient]
                object.
            initial_token_data: Token data to use for initial state, if not to be
                read from storage for initialization.
            token_file: A path to the storage location that should be used for
                credential storage. Credentials will be read from this location,
                and this location will be used to save refreshed credentials.
                Depending on the configuration, this may be
                ignored or may be None. When this is set to None,
                tokens will not be persisted to storage.
            profile_name: A profile name used to identify the Auth configration
                at runtime.
        """
        if token_file:
            token_file_path = pathlib.Path(token_file)
        else:
            token_file_path = None

        request_authenticator = auth_client.default_request_authenticator(credential=token_file_path)  # type: ignore
        if initial_token_data:
            request_authenticator.update_credential_data(initial_token_data)
        return Auth(
            auth_client=auth_client,
            request_authenticator=request_authenticator,
            token_file_path=token_file_path,
            profile_name=profile_name,
        )

    @staticmethod
    def initialize_from_config(
        client_config: AuthClientConfig,
        initial_token_data: Optional[dict] = None,
        token_file: Optional[Union[str, pathlib.PurePath]] = None,
        profile_name: Optional[str] = None,
    ) -> Auth:
        """
        Construct and initialize a working set of authentication primitives,
        returning them in a new container object.
        Parameters:
            client_config: A constructed AuthClientConfig object.
            initial_token_data: Token data to use for initial state, if not to be
                read from storage for initialization.
            token_file: A path to the storage location that should be used for
                credential storage. Credentials will be read from this location,
                and this location will be used to save refreshed credentials.
                Depending on the configuration, this may be
                ignored or may be None. When this is set to None,
                tokens will not be persisted to storage.
            profile_name: A profile name used to identify the Auth configration
                at runtime.
        """
        auth_client = AuthClient.from_config(config=client_config)
        return Auth.initialize_from_client(
            auth_client=auth_client,
            initial_token_data=initial_token_data,
            token_file=token_file,
            profile_name=profile_name,
        )

    @staticmethod
    def initialize_from_config_dict(
        client_config: dict,
        initial_token_data: Optional[dict] = None,
        token_file: Optional[Union[str, pathlib.PurePath]] = None,
        profile_name: Optional[str] = None,
        storage_provider: Optional[ObjectStorageProvider] = None,
    ) -> Auth:
        """
        Construct and initialize a working set of authentication primitives,
        returning them in a new container object.
        Parameters:
            client_config: A dictionary containing a valid auth client configuration.
            initial_token_data: Token data to use for initial state, if not to be
                read from storage for initialization.
            token_file: A path to a storage location that should be used for
                credential storage. Credentials will be read from this location,
                and this location may be used to save refreshed credentials.
                Depending on the configuration, this may be
                ignored or may be None. When this is set to None,
                tokens will not be persisted to storage.
            profile_name: A profile name used to identify the Auth configration
                at runtime.
            storage_provider: A custom storage provider to use.
        """
        auth_client_config = AuthClientConfig.from_dict(config_data=client_config)
        if storage_provider:
            auth_client_config.set_storage_provider(storage_provider=storage_provider)
        return Auth.initialize_from_config(
            client_config=auth_client_config,
            initial_token_data=initial_token_data,
            token_file=token_file,
            profile_name=profile_name,
        )



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from planet_auth.credential import Credential
from planet_auth.storage_utils import InvalidDataException, ObjectStorageProvider


class FileBackedApiKey(Credential):
    """
    Credential object for storing simple API key bearer tokens.
    """

    def __init__(
        self,
        api_key=None,
        prefix="Bearer",
        api_key_file=None,
        storage_provider: Optional[ObjectStorageProvider] = None,
    ):
        if api_key:
            init_data = {"api_key": api_key, "bearer_token_prefix": prefix}
        else:
            init_data = None

        super().__init__(data=init_data, file_path=api_key_file, storage_provider=storage_provider)

    def check_data(self, data):
        """
        Check that the supplied data represents a valid API key object.
        """
        super().check_data(data)
        if not data.get("api_key"):
            raise InvalidDataException(message="'api_key' not found in file " + str(self._file_path))
        if not data.get("bearer_token_prefix"):
            raise InvalidDataException(messae="'bearer_token_prefix' not found in file " + str(self._file_path))

    def api_key(self):
        """
        Get the current API Key.
        """
        return self.lazy_get("api_key")

    def bearer_token_prefix(self):
        """
        Get the bearer token prefix that is to be used with the API key.
        """
        return self.lazy_get("bearer_token_prefix")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import sys

from planet_auth import (
    AuthException,
    FileBackedOidcCredential,
    OidcAuthClient,
    ExpiredTokenException,
    ClientCredentialsAuthClientBase,
)

from .options import (
    opt_audience,
    opt_client_id,
    opt_client_secret,
    opt_extra,
    opt_human_readable,
    opt_open_browser,
    opt_organization,
    opt_password,
    opt_project,
    opt_refresh,
    opt_scope,
    opt_show_qr_code,
    opt_sops,
    opt_username,
    opt_yes_no,
)
from .util import recast_exceptions_to_click, post_login_cmd_helper, print_obj
from .jwt_cmd import json_dumps_for_jwt_dict, hazmat_print_jwt


def _check_client_type(ctx):
    if not isinstance(ctx.obj["AUTH"].auth_client(), OidcAuthClient):
        raise click.ClickException(
            f'"oauth" auth commands can only be used with OAuth type auth profiles.'
            f' The current profile "{ctx.obj["AUTH"].profile_name()}" is of type "{ctx.obj["AUTH"].auth_client()._auth_client_config.meta()["client_type"]}".'
        )


@click.group("oauth", invoke_without_command=True)
@click.pass_context
def cmd_oauth(ctx):
    """
    OAuth2 authentication commands.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)

    _check_client_type(ctx)


@cmd_oauth.command("login")
@opt_open_browser
@opt_show_qr_code
@opt_scope
@opt_audience()
@opt_organization
@opt_project
@opt_username()
@opt_password()
@opt_client_id
@opt_client_secret
@opt_sops
@opt_yes_no
@opt_extra
@click.pass_context
@recast_exceptions_to_click(AuthException, ValueError)
def cmd_oauth_login(
    ctx,
    scope,
    audience,
    open_browser,
    show_qr_code,
    organization,
    username,
    password,
    auth_client_id,
    auth_client_secret,
    sops,
    yes,
    project,
    extra,
):
    """
    Perform an initial login using OAuth.
    """
    login_extra = {}
    if extra:
        for extra_opt in extra:
            key, value = extra_opt.split("=", 1)
            login_extra[key] = value
    if project:
        # Planet Labs OAuth extension to request a token for a particular project
        login_extra["project_id"] = project
    if organization:
        # Used by Auth0's OAuth implementation to support their concept of selecting
        # a particular organization at login when the user belongs to more than one.
        login_extra["organization"] = organization

    current_auth_context = ctx.obj["AUTH"]
    current_auth_context.login(
        requested_scopes=scope,
        requested_audiences=audience,
        allow_open_browser=open_browser,
        allow_tty_prompt=True,
        display_qr_code=show_qr_code,
        username=username,
        password=password,
        client_id=auth_client_id,
        client_secret=auth_client_secret,
        extra=login_extra,
    )
    print("Login succeeded.")  # Errors should throw.
    post_login_cmd_helper(override_auth_context=current_auth_context, use_sops=sops, prompt_pre_selection=yes)


@cmd_oauth.command("refresh")
@opt_scope
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_refresh(ctx, scope):
    """
    Obtain a new credential using the saved refresh token.

    It is possible to request a refresh token with scopes that are different
    from what is currently possessed, but you will never be granted
    more scopes than what the user has authorized.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    if not saved_token.refresh_token():
        raise click.ClickException("No refresh_token found in " + str(saved_token.path()))

    saved_token.set_data(auth_client.refresh(saved_token.refresh_token(), scope).data())
    saved_token.save()


@cmd_oauth.command("list-scopes")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_list_scopes(ctx):
    """
    List available OAuth scopes.

    This command will query the auth server for available scopes that may be requested.
    """
    auth_client = ctx.obj["AUTH"].auth_client()
    available_scopes = auth_client.get_scopes()
    available_scopes.sort()
    if available_scopes:
        print_obj(available_scopes)
    else:
        print_obj([])


@cmd_oauth.command("discovery")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_discovery(ctx):
    """
    Look up OAuth server discovery information.
    """
    auth_client = ctx.obj["AUTH"].auth_client()
    discovery_json = auth_client.oidc_discovery()
    print_obj(discovery_json)


@cmd_oauth.command("validate-access-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_validate_access_token_remote(ctx, human_readable):
    """
    Validate the access token. Validation is performed by calling
    out to the auth provider's token introspection network service.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    validation_json = auth_client.validate_access_token_remote(saved_token.access_token())

    if not validation_json or not validation_json.get("active"):
        print_obj("INVALID")
        sys.exit(1)
    # print_obj(validation_json)
    print(json_dumps_for_jwt_dict(data=validation_json, human_readable=human_readable))


@cmd_oauth.command("validate-access-token-local")
@click.pass_context
@opt_audience()
@opt_scope
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_validate_access_token_local(ctx, audience, scope, human_readable):
    """
    Validate the access token locally.

    When scopes are passed to the access token validator, the validator
    performs an "any of" check.  It will assert that any one of the scopes
    is present in the token. The validator does not assert that all scopes
    are present.

    It no scopes are passed to the validator, none are required.

    NOTICE:
        This functionality is not supported for all OAuth implementations.
        Access tokens are intended for consumption by resource servers,
        and may be opaque to the client.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    # Throws on error.
    validation_json = auth_client.validate_access_token_local(
        access_token=saved_token.access_token(), required_audience=audience, scopes_anyof=scope
    )
    # print_obj(validation_json)
    print(json_dumps_for_jwt_dict(data=validation_json, human_readable=human_readable))


@cmd_oauth.command("validate-id-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_validate_id_token_remote(ctx, human_readable):
    """
    Validate the ID token. Validation is performed by calling
    out to the auth provider's token introspection network service.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    validation_json = auth_client.validate_id_token_remote(saved_token.id_token())

    if not validation_json or not validation_json.get("active"):
        print_obj("INVALID")
        sys.exit(1)
    # print_obj(validation_json)
    print(json_dumps_for_jwt_dict(data=validation_json, human_readable=human_readable))


@cmd_oauth.command("validate-id-token-local")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_validate_id_token_local(ctx, human_readable):
    """
    Validate the ID token. This command validates the ID token locally,
    checking the token signature and claims against expected values.
    While validation is performed locally, network access is still
    required to obtain the signing keys from the auth provider.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    # Throws on error.
    validation_json = auth_client.validate_id_token_local(saved_token.id_token())
    # print_obj(validation_json)
    print(json_dumps_for_jwt_dict(data=validation_json, human_readable=human_readable))


@cmd_oauth.command("validate-refresh-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_validate_refresh_token_remote(ctx, human_readable):
    """
    Validate the refresh token. Validation is performed by calling
    out to the auth provider's token introspection network service.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    validation_json = auth_client.validate_refresh_token_remote(saved_token.refresh_token())

    if not validation_json or not validation_json.get("active"):
        print_obj("INVALID")
        sys.exit(1)
    # print_obj(validation_json)
    print(json_dumps_for_jwt_dict(data=validation_json, human_readable=human_readable))


@cmd_oauth.command("revoke-access-token")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_revoke_access_token(ctx):
    """
    Revoke the current access token.

    Revoking the access token does not revoke the refresh token, which will
    remain powerful.

    It should be noted that while this command revokes the access token with
    the auth services, access tokens are bearer tokens, and may still be
    accepted by some service endpoints. It is up to each service whether
    access tokens are accepted as bearer tokens, or double verified against
    the auth services.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    auth_client.revoke_access_token(saved_token.access_token())


@cmd_oauth.command("revoke-refresh-token")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_revoke_refresh_token(ctx):
    """
    Revoke the current refresh token.

    After the refresh token has been revoked, it will be necessary to login
    again to access other services.  Revoking the refresh token does not
    revoke the current access token, which may remain potent until its
    natural expiration time if not also revoked.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    auth_client.revoke_refresh_token(saved_token.refresh_token())


@cmd_oauth.command("userinfo")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_userinfo(ctx):
    """
    Look up user information from the auth server using the access token.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    auth_client = ctx.obj["AUTH"].auth_client()
    saved_token.load()
    userinfo_json = auth_client.userinfo_from_access_token(saved_token.access_token())

    # print_obj("OK")
    print_obj(userinfo_json)


@cmd_oauth.command("print-access-token")
@click.pass_context
@opt_refresh
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_print_access_token(ctx, refresh):
    """
    Show the current OAuth access token.  Stale tokens will be automatically refreshed.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    saved_token.load()

    if refresh:
        auth_client = ctx.obj["AUTH"].auth_client()
        try:
            _ = auth_client.validate_access_token_local(access_token=saved_token.access_token())
        except ExpiredTokenException:
            # Client Credentials grant clients do not use a refresh token to refresh,
            # they re-login, which in their case is not user interactive.
            if saved_token.refresh_token():
                new_token = auth_client.refresh(saved_token.refresh_token())
            elif isinstance(auth_client, ClientCredentialsAuthClientBase):
                new_token = auth_client.login()
            else:
                raise click.ClickException("Cannot refresh expired token")  # pylint: disable=W0707

            saved_token.set_data(new_token.data())
            saved_token.save()

    # Not using object print for token printing. We don't want object quoting and escaping.
    # print_obj(saved_token.access_token())
    print(saved_token.access_token())


@cmd_oauth.command("decode-access-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_decode_jwt_access_token(ctx, human_readable):
    """
    Decode a JWT access token locally and display its contents.  NO
    VALIDATION IS PERFORMED.  This function is intended for local
    debugging purposes.  Note: Access tokens need not be JWTs.
    This function will not work for authorization servers that issue
    access tokens in other formats.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    hazmat_print_jwt(saved_token.access_token(), human_readable=human_readable)


@cmd_oauth.command("decode-id-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_decode_jwt_id_token(ctx, human_readable):
    """
    Decode a JWT ID token locally and display its contents.  NO
    VALIDATION IS PERFORMED.  This function is intended for local
    debugging purposes.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    hazmat_print_jwt(saved_token.id_token(), human_readable=human_readable)


@cmd_oauth.command("decode-refresh-token")
@click.pass_context
@opt_human_readable
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_oauth_decode_jwt_refresh_token(ctx, human_readable):
    """
    Decode a JWT refresh token locally and display its contents.  NO
    VALIDATION IS PERFORMED.  This function is intended for local
    debugging purposes.  Note: Refresh tokens need not be JWTs.
    This function will not work for authorization servers that issue
    refresh tokens in other formats.
    """
    saved_token = FileBackedOidcCredential(None, ctx.obj["AUTH"].token_file_path())
    hazmat_print_jwt(saved_token.refresh_token(), human_readable=human_readable)
